"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides availability special rules interfaces.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io//en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    MIT License

    Copyright (c) 2017, 2018, 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import socket

from urllib3 import exceptions as urllib3_exceptions

import PyFunceble


class ExtraRules:  # pylint: disable=too-few-public-methods # pragma: no cover
    """
    Manage some extra rules.,

    :param str subject: The subject we are working with.

    :param str subject_type:
        The type of the subject we are working with.
        Should be one of the following.

            - :code:`domain`
            - :code:`url`

    :param http_status_code: The extracted status code.
    :type http_status_code: str|int
    """

    def __init__(self, subject, subject_type, http_status_code):
        # We share the subject we are working with.
        self.subject = subject
        # We share the subject type.
        self.subject_type = subject_type
        # We share the status code.
        self.status_code = http_status_code

        user_agent = PyFunceble.engine.UserAgent().get()

        if user_agent:
            # We set the header that we will send when communicating with webservers.
            self.headers = {"User-Agent": user_agent}
        else:
            self.headers = {}

        # We set a list of regex and methods to call if matched.
        self.regexes_active_to_inactive_potentially_down = {
            r"\.blogspot\.": self.__blogspot,
            r"\.canalblog\.com$": self.__special_down_404,
            r"\.github\.io$": self.__special_down_404,
            r"\.liveadvert\.com$": self.__special_down_404,
            r"\.skyrock\.com$": self.__special_down_404,
            r"\.tumblr\.com$": self.__special_down_404,
        }

        # We set a list of regex and methods to call if matched.
        self.regexes_active_to_inactive_potentially_up = {
            r"\.blogspot\.": self.__blogspot,
            r"\.wordpress\.com$": self.__wordpress_dot_com,
        }

        PyFunceble.LOGGER.debug(f"[{self.subject}] Headers:\n{self.headers}")

    @classmethod
    def __special_down(cls):
        """
        Set what we return for the SPECIAL status de-escalation.

        :return: :code:`(new status, new source)`
        :rtype: tuple
        """

        return PyFunceble.STATUS.official.down, "SPECIAL"

    def __special_down_404(self):
        """
        Set what we return for the SPECIAL status de-escalation
        when the 404 status code is caught.

        :return: :code:`(new status, new source)`
        :rtype: tuple
        """

        if self.status_code == 404:
            return PyFunceble.STATUS.official.down, "SPECIAL"
        return None

    @classmethod
    def __special_up(cls):
        """
        Set what we return for the SPECIAL status escalation.

        :return: :code:`(new status, new source)`
        :rtype: tuple
        """

        return PyFunceble.STATUS.official.up, "SPECIAL"

    @classmethod
    def __http_status_code_up(cls):
        """
        Set what we return for the HTTP Code status escalation.

        :return: :code:`(new status, new source)`
        :rtype: tuple
        """

        return PyFunceble.STATUS.official.up, "HTTP Code"

    def __blogspot(self):
        """
        Handle the blogspot SPECIAL case.

        :return:
            :code:`(new status, new source)` or :code:`None` if there is any
            change to apply.
        :rtype: tuple|None
        """

        # We iniate a list of elements in the HTML which will tell us more about
        # the status of the domain.
        regex_blogger = ["create-blog.g?", "87065", "doesn&#8217;t&nbsp;exist"]

        if self.subject_type in ["domain", "file_domain"]:
            # The element we are testing is a domain.

            # We construct the url to get.
            url_to_get = "http://%s" % self.subject
        elif self.subject_type in ["url", "file_url"]:
            # The element we are testing is a URL.

            # We construct the url to get.
            url_to_get = self.subject
        else:
            raise ValueError("Given subject type not registered.")

        try:
            # We get the HTML of the home page.
            blogger_content_request = PyFunceble.REQUESTS.get(
                url_to_get,
                headers=self.headers,
                timeout=PyFunceble.CONFIGURATION.timeout,
                verify=PyFunceble.CONFIGURATION.verify_ssl_certificate,
                allow_redirects=True,
            )

            for regx in regex_blogger:
                # We loop through the list of regex to match.

                if regx in blogger_content_request.text or PyFunceble.helpers.Regex(
                    regx
                ).match(blogger_content_request.text, return_match=False):
                    # * The currently read regex is present into the docuement.
                    # or
                    # * Something in the document match the currently read regex.

                    PyFunceble.LOGGER.info(
                        "[{self.subject}] Switching status according to blogspot rule."
                    )

                    # We update the status and source.
                    return self.__special_down()
        except (
            PyFunceble.REQUESTS.exceptions.InvalidURL,
            socket.timeout,
            PyFunceble.REQUESTS.exceptions.Timeout,
            PyFunceble.REQUESTS.exceptions.ConnectionError,
            urllib3_exceptions.InvalidHeader,
            UnicodeDecodeError,  # The probability that this happend in production is minimal.
        ):
            PyFunceble.LOGGER.exception()

        # We return None, there is no changes.
        return None

    def __wordpress_dot_com(self):
        """
        Handle the wordpress.com SPECIAL case.

        :return:
            :code:`(new status, new source)` or :code:`None` if there is any
            change to apply.
        :rtype: tuple|None
        """

        # We initiate a variable which whill have to be into the HTML
        # in order to be considered as inactive.
        does_not_exist = "doesn&#8217;t&nbsp;exist"

        try:
            # We get the content of the page.
            wordpress_com_content = PyFunceble.REQUESTS.get(
                "http://{}:80".format(self.subject),
                headers=self.headers,
                timeout=PyFunceble.CONFIGURATION.timeout,
                verify=PyFunceble.CONFIGURATION.verify_ssl_certificate,
                allow_redirects=True,
            )

            if does_not_exist in wordpress_com_content.text:
                # The marker is into the page content.

                PyFunceble.LOGGER.info(
                    "[{self.subject}] Switching status according to wordpress_dot_com rule."
                )

                # We return the new status and source.
                return self.__special_down()
        except PyFunceble.REQUESTS.exceptions.SSLError:
            pass

        # We return None, there is no changes.
        return None

    def __handle_potentially_inactive(self, previous_state):
        """
        Handle the potentially inactive case.

        :param str previous_state: The previously catched status.

        :return:
            :code:`(new status, new source)` or :code:`None` if there is any
            change to apply.
        :rtype: tuple|None
        """

        if (
            PyFunceble.HTTP_CODE.active
            and self.status_code in PyFunceble.HTTP_CODE.list.potentially_down
        ):
            # * The http status request is activated.
            # and
            # * The extracted http status code is in the list of
            #   potentially down list.

            # We generate the analytics files.
            PyFunceble.output.Generate(
                self.subject, self.subject_type, previous_state
            ).analytic_file("potentially_down")

            if not PyFunceble.CONFIGURATION.no_special:
                # We are authorized to play with the SPEICIAL rules.

                for regx in self.regexes_active_to_inactive_potentially_down:
                    # We loop through the list of available regex.

                    if PyFunceble.helpers.Regex(regx).match(
                        self.subject, return_match=False
                    ):
                        # The element we are currently testing match the
                        # regex we are currently reading.

                        # We get the output of the function associated
                        # with the regex.
                        output = self.regexes_active_to_inactive_potentially_down[
                            regx
                        ]()

                        if output is not None:
                            # The output is not None.

                            # We return the new source and state.
                            return output

        # We return None, there is no changes.
        return None

    def __handle_potentially_up(self):
        """
        Handle the potentially up  case.

        :return:
            :code:`(new status, new source)` or :code:`None` if there is any
            change to apply.
        :rtype: tuple|None
        """

        if (
            PyFunceble.HTTP_CODE.active
            and self.status_code in PyFunceble.HTTP_CODE.list.potentially_up
        ):
            # * The http status code request is activated.
            # and
            # * The extracted http status code is into the list of potentially up codes.

            if not PyFunceble.CONFIGURATION.no_special:
                # We are authorized to play with the SPEICIAL rules.

                for regx in self.regexes_active_to_inactive_potentially_up:
                    # We loop through the list of available regex.

                    if PyFunceble.helpers.Regex(regex=regx).match(
                        self.subject, return_match=False
                    ):
                        # The element we are currently testing match the
                        # regex we are currently reading.

                        # We get the output of the function associated
                        # with the regex.
                        output = self.regexes_active_to_inactive_potentially_up[regx]()

                        if output is not None:
                            # The output is not None.

                            # We return the new source and state.
                            return output

        # We return None, there is no changes.
        return None

    def __handle_http_code(self, previous_state):
        """
        Handle the HTTP Code status escalation.

        :param str previous_state: The previously catched status.

        :return:
            :code:`(new status, new source)` or :code:`None` if there is any
            change to apply.
        :rtype: tuple|None
        """

        try:

            if self.status_code in PyFunceble.HTTP_CODE.list.up:
                # The extracted http code is in the list of up codes.

                # We generate the analytics files.
                PyFunceble.output.Generate(
                    self.subject, self.subject_type, previous_state
                ).analytic_file(PyFunceble.STATUS.official.up)

                if previous_state.lower() not in PyFunceble.STATUS.list.up:
                    # And we return the new status and source

                    PyFunceble.LOGGER.info(
                        "[{self.subject}] Switching status according to status code rule."
                    )

                    return self.__http_status_code_up()

            if self.status_code in PyFunceble.HTTP_CODE.list.potentially_up:
                # The extracted http status code is in the list of potentially up status.

                # We generate the analytics files.
                PyFunceble.output.Generate(
                    self.subject, self.subject_type, previous_state
                ).analytic_file("potentially_up")

                if previous_state.lower() not in PyFunceble.STATUS.list.up:

                    PyFunceble.LOGGER.info(
                        "[{self.subject}] Switching status according to status code rule."
                    )

                    # And we return the new status and source
                    return self.__http_status_code_up()

            if (
                previous_state.lower() in PyFunceble.STATUS.list.down
                and self.status_code != PyFunceble.HTTP_CODE.not_found_default
            ):
                # We generate the analytics files.
                PyFunceble.output.Generate(
                    self.subject, self.subject_type, previous_state
                ).analytic_file("potentially_up")

                PyFunceble.LOGGER.info(
                    "[{self.subject}] Switching status according to status code rule."
                )

                # And we return the new status and source
                return self.__http_status_code_up()

            if (
                previous_state.lower() not in PyFunceble.STATUS.list.down
                or previous_state.lower() not in PyFunceble.STATUS.list.invalid
            ) and self.status_code in PyFunceble.HTTP_CODE.list.potentially_down:
                # The extracted http code is in the list of potentially down status code.

                # We generate the analytics files.
                PyFunceble.output.Generate(
                    self.subject, self.subject_type, previous_state
                ).analytic_file("potentially_down")
        except KeyError:
            PyFunceble.LOGGER.exception()

        # We return None, there is no changes.
        return None

    def __handle_reputation(self):
        """
        Handle the reputation escalation.

        :return:
            :code:`(new status, new source)` or :code:`None` if there is any
            change to apply.
        :rtype: tuple|None
        """

        if (
            not PyFunceble.CONFIGURATION.no_special
            and PyFunceble.CONFIGURATION.use_reputation_data
            and self.subject in PyFunceble.lookup.IPv4Reputation()
        ):
            PyFunceble.LOGGER.info(
                "[{self.subject}] Switching status according to reputation rule."
            )

            return self.__special_up()
        return None

    def __handle_ip_range(self):
        """
        Handle the IP range status escalation.

        :return:
            :code:`(new status, new source)` or :code:`None` if there is any
            change to apply.
        :rtype: tuple|None
        """

        if (
            not PyFunceble.CONFIGURATION.no_special
            and PyFunceble.Check(self.subject).is_ip_range()
        ):
            # * We can run/check the special rule.
            # and
            # * The element we are currently testing is an IP range.

            PyFunceble.LOGGER.info(
                "[{self.subject}] Switching status according to IP range rule."
            )

            # We return the new status and source.
            return self.__special_up()

        # We return None, there is no changes.
        return None

    def handle(
        self, previous_state, previous_source
    ):  # pylint:disable= too-many-return-statements
        """
        Globally handle the case of the currently tested domain.
        """

        # We preset the new status and the source to None.
        new_status = None
        source = None

        # We convert the given previous state to lower case.
        previous_state_modified = previous_state.lower()

        if previous_state_modified in PyFunceble.STATUS.list.up:
            # The previous state is in the list of up status.

            try:
                # We try to get the new status and source from another handler.

                new_status, source = self.__handle_potentially_inactive(previous_state)

                return new_status, source
            except TypeError:
                pass

            try:
                # We try to get the new status and source from another handler.

                new_status, source = self.__handle_potentially_up()
                return new_status, source
            except TypeError:
                pass

        if previous_state_modified in PyFunceble.STATUS.list.valid:
            # The previous state is in the list of valid status.

            # We return the given state and source, nothing changes.
            return previous_state, previous_source

        if previous_state_modified in PyFunceble.STATUS.list.down:
            # The previous state is in the list of down status.

            try:
                # We try to get the new status and source from another handler.

                new_status, source = self.__handle_ip_range()

                return new_status, source
            except TypeError:
                pass

            if PyFunceble.HTTP_CODE.active:
                # The http status code request is activated.

                try:
                    # We try to get the new status and source from another handler.

                    new_status, source = self.__handle_http_code(previous_state)

                    return new_status, source
                except TypeError:
                    pass

        if previous_state_modified in PyFunceble.STATUS.list.invalid:
            # The previous state is in the list of invalid status.

            if PyFunceble.HTTP_CODE.active:
                # The http status code request is activated.

                try:
                    # We try to get the new status and source from another handler.

                    new_status, source = self.__handle_http_code(previous_state)

                    return new_status, source
                except TypeError:
                    pass

        # We return the given state and source, nothing changes.
        return previous_state, previous_source



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the inactive database interface.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io//en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    MIT License

    Copyright (c) 2017, 2018, 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

from datetime import datetime, timedelta

import PyFunceble


class InactiveDB:  # pylint: disable=too-many-instance-attributes
    """
    Provides the inactive database logic and interface.

    :param str filename: The name of the file we are processing.
    """

    # We initiate a variable which will save the
    # cache about who are present or not into the database.
    is_present_cache = {}

    # Saves the whole database content.
    database = {}

    # Save the operation authorization.
    authorized = False

    # Save the filename we are operating.
    filename = None

    def __init__(self, filename, mysql_db=None, parent_process=False):
        self.one_day = timedelta(days=1)
        self.database_file = ""

        # We get the authorization status.
        self.authorized = self.authorization()
        # We share the parent state.
        self.parent = parent_process

        PyFunceble.LOGGER.debug(f"Authorization: {self.authorized}")

        if self.authorized:
            # We convert the number of days between the database retest
            # to seconds.
            self.days = timedelta(days=PyFunceble.CONFIGURATION.days_between_db_retest)

            if PyFunceble.CONFIGURATION.db_type == "json":
                # We set the path to the inactive database file.
                self.database_file = "{0}{1}".format(
                    PyFunceble.CONFIG_DIRECTORY,
                    PyFunceble.OUTPUTS.default_files.inactive_db,
                )

            # We share the filename.
            self.filename = filename

            # We get the db instance.
            self.mysql_db = mysql_db

            self.table_name = self.get_table_name()
            self.to_retest = self.get_to_retest()

            PyFunceble.LOGGER.debug(f"DB: {self.mysql_db}")
            PyFunceble.LOGGER.debug(f"Table Name: {self.table_name}")
            PyFunceble.LOGGER.debug(f"DB (File): {self.database_file}")

            # We initiate the database.
            self.initiate()

    def __contains__(self, subject):
        if self.authorized:
            # We are authorized to operate.

            if PyFunceble.CONFIGURATION.db_type == "json":
                if subject not in self.is_present_cache:
                    self.is_present_cache[subject] = False

                    for element in [
                        x for x in self.database[self.filename].keys() if x.isdigit()
                    ]:
                        if subject in self[element]:
                            self.is_present_cache[subject] = True
                            break

                return self.is_present_cache[subject]

            if PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
                query = (
                    "SELECT COUNT(*) "
                    "FROM {0} "
                    "WHERE subject = %(subject)s AND file_path = %(file)s"
                ).format(self.table_name)

                with self.mysql_db.get_connection() as cursor:
                    cursor.execute(query, {"subject": subject, "file": self.filename})

                    fetched = cursor.fetchone()

                return fetched["COUNT(*)"]

        return False  # pragma: no cover

    def __getitem__(self, index):
        if (
            self.authorized
            and PyFunceble.CONFIGURATION.db_type == "json"
            and self.filename in self.database
            and index in self.database[self.filename]
        ):
            return self.database[self.filename][index]
        return []

    def __setitem__(self, index, value):
        if PyFunceble.CONFIGURATION.db_type == "json":
            actual_state = self[index]

            if actual_state:
                if isinstance(actual_state, dict):
                    if isinstance(value, dict):  # pragma: no cover
                        self.database[self.filename][index] = PyFunceble.helpers.Merge(
                            value
                        ).into(self.database[self.filename][index])
                    else:  # pragma: no cover
                        self.database[self.filename][index] = value
                elif isinstance(actual_state, list):  # pragma: no cover
                    if isinstance(value, list):
                        PyFunceble.helpers.Merge(value).into(
                            self.database[self.filename][index], strict=False
                        )
                    else:  # pragma: no cover
                        self.database[self.filename][index].append(value)

                    self.database[self.filename][index] = PyFunceble.helpers.List(
                        self.database[self.filename][index]
                    ).format()
                else:  # pragma: no cover
                    self.database[self.filename][index] = value
            else:
                if self.filename not in self.database:
                    self.database[self.filename] = {index: value}
                else:
                    self.database[self.filename][index] = value

    @classmethod
    def authorization(cls):
        """
        Provides the execution authorization.
        """

        return PyFunceble.CONFIGURATION.inactive_database

    def get_table_name(self):
        """
        Returns the name of the table to use.
        """

        if PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
            return self.mysql_db.tables["inactive"]
        return "inactive"

    def _merge(self):
        """
        Merges the database with the older one which
        has already been set into the database.
        """

        if self.authorized and PyFunceble.CONFIGURATION.db_type == "json":
            # We are authorized to operate.

            # We get the content of the database.
            database_content = PyFunceble.helpers.Dict().from_json_file(
                self.database_file
            )

            # We get the database top keys.
            database_top_keys = [
                x for x in database_content.keys() if database_content[x]
            ]

            for database_top_key in database_top_keys:
                # We loop through the list of database top keys.

                # We get the list of lower indexes.
                database_low_keys = database_content[database_top_key].keys()

                for database_low_key in database_low_keys:
                    # We loop through the lower keys.

                    if isinstance(
                        database_content[database_top_key][database_low_key], list
                    ):  # pragma: no cover
                        to_set = {
                            x: ""
                            for x in database_content[database_top_key][
                                database_low_key
                            ]
                        }
                    else:
                        to_set = database_content[database_top_key][database_low_key]

                    if database_top_key not in self.database:
                        self.database[database_top_key] = {database_low_key: to_set}
                    else:
                        if database_low_key in self.database[database_top_key]:
                            self.database[database_top_key][
                                database_low_key
                            ] = PyFunceble.helpers.Merge(to_set).into(
                                self.database[database_top_key][database_low_key],
                                strict=False,
                            )
                        else:  # pragma: no cover
                            self.database[database_top_key][database_low_key] = to_set

            PyFunceble.LOGGER.info("Merged possible old to the new format")

    def load(self):
        """
        Loads the content of the database file.
        """

        if self.authorized and PyFunceble.CONFIGURATION.db_type == "json":
            # We are authorized to operate.

            if PyFunceble.helpers.File(self.database_file).exists():
                # The database file exists.

                self._merge()

                if (
                    self.filename in self.database
                    and "to_test" in self.database[self.filename]
                ):
                    new_time = str(
                        int(
                            (
                                datetime.now() - self.one_day - timedelta(seconds=100)
                            ).timestamp()
                        )
                    )
                    self.database[self.filename][new_time] = self.database[
                        self.filename
                    ]["to_test"]

                    del self.database[self.filename]["to_test"]
            else:
                # The database file do not exists.

                # We initiate an empty database.
                self.database = {self.filename: {}}

            if self.filename not in self.database:  # pragma: no cover
                # We create the current file namepace
                self.database[self.filename] = {}

            PyFunceble.LOGGER.info(
                "Database content loaded in memory. (DATASET WONT BE LOGGED)"
            )

    def save(self):
        """
        Saves the current database into the database file.
        """

        if (
            self.authorized
            and self.parent
            and PyFunceble.CONFIGURATION.db_type == "json"
        ):
            # We are authorized to operate.

            # We save the current database state into the database file.
            PyFunceble.helpers.Dict(self.database).to_json_file(self.database_file)

            PyFunceble.LOGGER.info(f"Saved database into {repr(self.database_file)}.")

    def initiate(self):
        """
        Initiates the databsse.
        """

        if self.authorized:
            # * We are authorized to operate.
            # and
            # * The filename is already in the database.

            # We load the database.
            self.load()

            self.save()

    @classmethod
    def timestamp(cls):
        """
        Gets the timestamp where we are going to save our current list.

        :return: The timestamp to append with the currently tested element.
        :rtype: int|str
        """

        # We return the current time.
        return int(datetime.now().timestamp())

    def add(self, subject, status):
        """
        Adds the given subject into the database.

        :param str subject: The subject we are working with.
        :param str status: The status of the given subject.
        """

        if self.authorized:
            # We are authorized to operate.

            if PyFunceble.CONFIGURATION.db_type == "json":
                # We get the timestamp to use as index.
                timestamp = str(self.timestamp())

                if self.filename in self.database:
                    # * The file path is not into the database.

                    self.remove(subject)

                # We initiate the file path and its content into the database.
                self[timestamp] = {subject: status}

                PyFunceble.LOGGER.info(
                    f"Indexed {repr(subject)} with the status "
                    f"{repr(status)} into {repr(self.filename)} database's."
                )

                # And we save the database.
                self.save()
            elif PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
                digest = PyFunceble.helpers.Hash(algo="sha256").data(
                    bytes(self.filename + subject, "utf-8")
                )

                query = (
                    "INSERT INTO {0} "
                    "(file_path, subject, status, digest) "
                    "VALUES (%(file)s, %(subject)s, %(status)s, %(digest)s)"
                ).format(self.table_name)

                with self.mysql_db.get_connection() as cursor:

                    playload = {
                        "file": self.filename,
                        "subject": subject,
                        "status": status,
                        "digest": digest,
                    }

                    try:
                        cursor.execute(query, playload)

                        PyFunceble.LOGGER.info(
                            f"Inserted into the database: \n {playload}"
                        )
                    except self.mysql_db.errors:
                        query = (
                            "UPDATE {0} "
                            "SET subject = %(subject)s, status = %(status)s "
                            "WHERE digest = %(digest)s"
                        ).format(self.table_name)

                        cursor.execute(
                            query,
                            {"subject": subject, "status": status, "digest": digest},
                        )

                        PyFunceble.LOGGER.info(
                            "Data already indexed, updated the modified "
                            f"column of the row related to {repr(subject)}."
                        )

    def remove(self, subject):
        """
        Removes all occurrences of the given subject from the database.

        :param str subject: The subject we are working with.
        """

        if self.authorized:
            # We are authorized to operate.

            if PyFunceble.CONFIGURATION.db_type == "json":
                for data in self.database[self.filename]:
                    # We loop through the index of the file database.

                    if subject in self[data]:
                        # The currently tested element into the currently read index.

                        self[data] = PyFunceble.helpers.Dict(self[data]).remove_key(
                            subject
                        )

                        PyFunceble.LOGGER.info(
                            "Cleaned the data related to " f"{repr(subject)}."
                        )

                to_delete = [
                    x for x, y in self.database[self.filename].items() if not y
                ]

                for index in to_delete:
                    del self.database[self.filename][index]

                # And we save the data into the database.
                self.save()
            elif PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
                # We construct the query we are going to execute.
                query = (
                    "DELETE FROM {0} "
                    "WHERE file_path = %(file)s "
                    "AND subject = %(subject)s"
                ).format(self.table_name)

                with self.mysql_db.get_connection() as cursor:
                    cursor.execute(query, {"file": self.filename, "subject": subject})

                    PyFunceble.LOGGER.info(
                        "Cleaned the data related to "
                        f"{repr(subject)} and {repr(self.filename)} from "
                        "the {repr(self.table_name)} table."
                    )

    def __execute_query_retest_already_tested(self, query):  # pragma: no cover
        """
        Executes the query to get the list to retest or already tested.
        """

        with self.mysql_db.get_connection() as cursor:
            cursor.execute(query, {"file": self.filename, "days": self.days})
            fetched = cursor.fetchall()

            if fetched:
                return {x["subject"] for x in fetched}

        return set()

    def get_to_retest(self):  # pylint: pragma: no cover
        """
        Returns a set of subject to restest.
        """

        PyFunceble.LOGGER.info(
            "Getting the list of subjects to retest (DATASET WONT BE LOGGED)"
        )

        if self.authorized:
            if PyFunceble.CONFIGURATION.db_type == "json":
                try:
                    return {
                        z
                        for x, y in self.database[self.filename].items()
                        if x.isdigit()
                        and datetime.now()
                        > datetime.fromtimestamp(float(x)) + self.days
                        for z in y.keys()
                    }
                except KeyError:
                    return set()

            if PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
                query = (
                    "SELECT * FROM {0} WHERE file_path = %(file)s "
                    "AND CAST(UNIX_TIMESTAMP() AS {1}) "
                    "> (CAST(UNIX_TIMESTAMP(modified) AS {1}) + CAST(%(days)s AS {1}))"
                ).format(self.table_name, self.mysql_db.int_cast_type)

                return self.__execute_query_retest_already_tested(query)
        return set()

    def get_already_tested(self):  # pragma: no cover
        """
        Returns a set of already tested subjects.
        """

        PyFunceble.LOGGER.info(
            "Getting the list of already tested (DATASET WONT BE LOGGED)"
        )

        if self.authorized:
            if PyFunceble.CONFIGURATION.db_type == "json":
                try:
                    return {
                        z
                        for x, y in self.database[self.filename].items()
                        if x.isdigit()
                        and datetime.now()
                        < datetime.fromtimestamp(float(x)) + self.days
                        for z in y.keys()
                    }
                except KeyError:
                    return set()

            if PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
                query = (
                    "SELECT * FROM {0} WHERE file_path= %(file)s "
                    "AND CAST(UNIX_TIMESTAMP() AS {1}) "
                    "< (CAST(UNIX_TIMESTAMP(modified) AS {1}) + CAST(%(days)s AS {1}))"
                ).format(self.table_name, self.mysql_db.int_cast_type)

                return self.__execute_query_retest_already_tested(query)
        return set()



# pylint:disable=line-too-long
"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Tests of PyFunceble.extractor.expiration_date

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io//en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    MIT License

    Copyright (c) 2017, 2018, 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""
# pylint: enable=line-too-long

from random import choice
from unittest import TestCase
from unittest import main as launch_tests

from PyFunceble.extractor import ExpirationDate


class TestExpirationDate(TestCase):
    """
    Tests of PyFunceble.engine.expiration_date.
    """

    def setUp(self):
        """
        Setups everything needed for the tests.
        """

        self.dates_samples = [
            "02-jan-2017",
            "02.01.2017",
            "02/01/2017",
            "2017-01-02",
            "2017.01.02",
            "2017/01/02",
            "2017.01.02 15:00:00",
            "20170102 15:00:00",
            "2017-01-02 15:00:00",
            "02.01.2017 15:00:00",
            "02-Jan-2017 15:00:00 UTC",
            "2017/01/02 01:00:00 (+0900)",
            "2017/01/02 01:00:00",
            "Mon Jan 02 15:00:00 GMT 2017",
            "Mon Jan 02 2017",
            "2017-01-02T15:00:00",
            "2017-01-02T15:00:00Z",
            "2017-01-02T15:00:00+0200",
            "2017-01-02T15:00:00+0200.622265+03:00",
            "2017-01-02T15:00:00+0200.622265",
            "2017-01-02T23:59:59.0Z",
            "02-01-2017",
            "2017. 01. 02.",
            "2017-01-02T00:00:00+13:00",
            "20170102",
            "02-Jan-2017",
            "02.1.2017",
            "02 Jan 2017",
            "02-January-2017",
            "2017-Jan-02.",
            "Mon Jan 02 15:00:00 2017",
            "January 02 2017-Jan-02",
            "2.1.2017",
            "20170102000000",
            "January  2 2017",
            "2nd January 2017",
        ]

        self.expiration_date_markers = [
            "expire: ",
            "expire on: ",
            "Expiry Date: ",
            "free-date ",
            "expires: ",
            "Expiration date: ",
            "Expiry date: ",
            "Expire Date: ",
            "renewal date: ",
            "Expires: ",
            "validity: ",
            "Expiration Date             : ",
            "Expiry : ",
            "expires at: ",
            "domain_datebilleduntil: ",
            "Data de expiração / Expiration Date (dd/mm/yyyy): ",
            "Fecha de expiración (Expiration date): ",
            "[Expires on] ",
            "status:      OK-UNTIL ",
            "renewal: ",
            "expires............: ",
            "expire-date: ",
            "Exp date: ",
            "Valid-date ",
            "Expires On: ",
            "Fecha de vencimiento: ",
            "Expiration:......... ",
            "Fecha de Vencimiento: ",
            "Registry Expiry Date: ",
            "Expires on..............: ",
            "Expiration Time: ",
            "Expiration Date: ",
            "Expired: ",
            "Date d'expiration: ",
            "expiration date: ",
        ]

        self.expiration_dates = []

        for marker in self.expiration_date_markers:
            self.expiration_dates.append(f"{marker} {choice(self.dates_samples)}")

    def test_extractor(self):
        """
        Tests extractor.
        """

        expected = "02-jan-2017"

        for date in self.expiration_dates:
            actual = ExpirationDate(date).get_extracted()

            self.assertEqual(expected, actual, date)

        actual = (
            ExpirationDate(
                f"Record expires on {choice(self.dates_samples)} (YYYY-MM-DD)"
            ).get_extracted(),
        )

        expected = None
        actual = ExpirationDate("").get_extracted()

        self.assertEqual(expected, actual)


if __name__ == "__main__":
    launch_tests()



# pylint:disable=line-too-long
"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Tests of PyFunceble.helpers.merge

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io//en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    MIT License

    Copyright (c) 2017, 2018, 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""
# pylint: enable=line-too-long
from unittest import TestCase
from unittest import main as launch_tests

from PyFunceble.helpers.merge import Merge


class TestMerge(TestCase):
    """
    Tests of PyFunceble.helpers.merge
    """

    def test_str(self):
        """
        Tests the case that we want to merge 2 strs.
        """

        expected = "Hello"
        actual = Merge("Hello").into("World")

        self.assertEqual(expected, actual)

    def test_int(self):
        """
        Tests the case that we want to merge 2 ints.
        """

        expected = 2
        actual = Merge(2).into(1)

        self.assertEqual(expected, actual)

    def test_float(self):
        """
        Tests the case that we want to merge 2 floats.
        """

        expected = 2.1
        actual = Merge(2.1).into(3.1)

        self.assertEqual(expected, actual)

    def test_strict_simple_dict(self):
        """
        Tests the case that we want to strictly merge 2 simple dict.
        """

        given = {
            "world": "Fun Ilrys",
            "hello_world": {"author": "funilrys", "name": "Fun"},
            "hello": ["This is PyFunceble!", "Uhh!"],
        }
        to_merge = {
            "hello_world": {"author": "nobody", "surname": "body"},
            "hello": ["hello", "Uhh"],
        }
        expected = {
            "hello_world": {"author": "nobody", "name": "Fun", "surname": "body"},
            "world": "Fun Ilrys",
            "hello": ["hello", "Uhh"],
        }
        actual = Merge(to_merge).into(given, strict=True)

        self.assertEqual(expected, actual)

    def test_not_strict_simple_dict(self):
        """
        Tests the case that we want to merge 2 simple dicts.
        """

        given = {
            "hello": ["This is PyFunceble!", "Uhh!"],
            "world": "Fun Ilrys",
            "hello_world": {"author": "funilrys", "name": "Fun"},
        }
        to_merge = {
            "hello": ["hello", "Uhh"],
            "hello_world": {"author": "nobody", "surname": "body"},
        }
        expected = {
            "world": "Fun Ilrys",
            "hello_world": {"author": "nobody", "name": "Fun", "surname": "body"},
            "hello": ["This is PyFunceble!", "Uhh!", "hello", "Uhh"],
        }

        actual = Merge(to_merge).into(given, strict=False)

        self.assertEqual(expected, actual)

    def test_strict_dict(self):
        """
        Tests the case that we want to strictly merge 2 dicts.
        """

        given = {
            "hello": {"world": ["This is PyFunceble!", "Uhh!"]},
            "world": "Fun Ilrys",
            "hello_world": {"author": "funilrys", "name": "Fun"},
        }
        to_merge = {
            "hello": {"world": ["hello", "Uhh!"]},
            "hello_world": {"author": "nobody", "surname": "body"},
        }
        expected = {
            "hello": {"world": ["hello", "Uhh!"]},
            "world": "Fun Ilrys",
            "hello_world": {"author": "nobody", "name": "Fun", "surname": "body"},
        }

        actual = Merge(to_merge).into(given, strict=True)

        self.assertEqual(expected, actual)

    def test_not_strict_dict(self):
        """
        Tests the case that we want to merge 2 dicts.
        """

        given = {
            "hello": {"world": ["This is PyFunceble!", "Uhh!"]},
            "world": "Fun Ilrys",
            "hello_world": {"author": "funilrys", "name": "Fun"},
        }
        to_merge = {
            "hello": {"world": ["hello", "Uhh!"]},
            "hello_world": {"author": "nobody", "surname": "body"},
        }
        expected = {
            "hello": {"world": ["This is PyFunceble!", "Uhh!", "hello"]},
            "world": "Fun Ilrys",
            "hello_world": {"author": "nobody", "name": "Fun", "surname": "body"},
        }

        actual = Merge(to_merge).into(given, strict=False)

        self.assertEqual(expected, actual)

    def test_strict_simple_list(self):
        """
        Tests the case that we want to strictly merge 2 simple lists.
        """

        given = [1, 2, 3, 4]
        to_merge = [2, 4, 5, 6, 7]

        expected = [2, 4, 5, 6, 7]
        actual = Merge(to_merge).into(given, strict=True)

        self.assertEqual(actual, expected)

    def test_not_strict_simple_list(self):
        """
        Tests the case that we want to merge 2 simple lists.
        """

        given = [1, 2, 3, 4]
        to_merge = [2, 4, 5, 6, 7]

        expected = [1, 2, 3, 4, 5, 6, 7]
        actual = Merge(to_merge).into(given, strict=False)

        self.assertEqual(actual, expected)

    def test_strict_list(self):
        """
        Tests the case that we want to strictly merge 2 lists.
        """

        given = ["hello", "world", 5, {"hello": "world"}, [1, 2, 3]]
        to_merge = ["hello", "world", 5, {"world": "hello"}, [4, 5]]
        expected = ["hello", "world", 5, {"hello": "world", "world": "hello"}, [4, 5]]

        actual = Merge(to_merge).into(given, strict=True)

        self.assertEqual(expected, actual)

        to_merge = ["hello", "world", 5, {"hello": "you!"}, [1, 2, 4, 5]]
        expected = ["hello", "world", 5, {"hello": "you!"}, [1, 2, 4, 5]]

        actual = Merge(to_merge).into(given, strict=True)

        self.assertEqual(expected, actual)

        to_merge = ["hello", "world", 5, {"hello": "you!"}, [1, 2, 4, 5]]
        expected = ["hello", "world", 5, {"hello": "you!"}, [1, 2, 4, 5]]

        actual = Merge(to_merge).into(given, strict=True)

        self.assertEqual(expected, actual)

    def test_not_strict_list(self):
        """
        Tests the case that we want to merge 2 lists.
        """

        given = ["hello", "world", 5, {"hello": "world"}, [1, 2, 3]]
        to_merge = ["hello", "world", 5, {"world": "hello"}, [4, 5]]
        expected = [
            "hello",
            "world",
            5,
            {"hello": "world"},
            [1, 2, 3],
            {"world": "hello"},
            [4, 5],
        ]

        actual = Merge(to_merge).into(given, strict=False)

        self.assertEqual(expected, actual)

        to_merge = ["hello", "world", 5, {"hello": "you!"}, [1, 2, 4, 5]]
        expected = [
            "hello",
            "world",
            5,
            {"hello": "world"},
            [1, 2, 3],
            {"hello": "you!"},
            [1, 2, 4, 5],
        ]

        actual = Merge(to_merge).into(given, strict=False)

        self.assertEqual(expected, actual)

        to_merge = ["hello", "world", 5, {"hello": "you!"}, [1, 2, 4, 5]]
        expected = [
            "hello",
            "world",
            5,
            {"hello": "world"},
            [1, 2, 3],
            {"hello": "you!"},
            [1, 2, 4, 5],
        ]

        actual = Merge(to_merge).into(given, strict=False)

        self.assertEqual(expected, actual)

    def test_mixed_str_int(self):
        """
        Tests the case that we want to merge a str into an int.
        """

        expected = "Hello"
        actual = Merge("Hello").into(1)

        self.assertEqual(expected, actual)

    def test_mixed_int_float(self):
        """
        Tests the case that we want to merge an int into a float.
        """

        expected = 1
        actual = Merge(1).into(2.1)

        self.assertEqual(expected, actual)


if __name__ == "__main__":
    launch_tests()



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the merging helpers.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io//en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    MIT License

    Copyright (c) 2017, 2018, 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""


class Merge:
    """
    Simplify the merging of dict and list.

    :param main: The main data to work with.
    :type main: str, tuple, bool, int, dict, list, float
    """

    def __init__(self, main):
        self.main = main

    def __list(self, origin, strict=True):
        """
        Process the list merging.

        :param bool strict:
            Activates the strict mode.

        :rtype: list
        """

        result = []

        if strict:
            for index, element in enumerate(self.main):
                try:
                    if isinstance(element, dict) and isinstance(origin[index], dict):
                        result.append(Merge(element).into(origin[index], strict=strict))
                    elif isinstance(element, list) and isinstance(origin[index], list):
                        result.append(Merge(element).into(origin[index], strict=strict))
                    else:
                        result.append(element)
                except IndexError:  # pragma: no cover
                    result.append(element)
        else:
            result = origin

            for element in self.main:
                if element not in result:
                    result.append(element)

        return result

    def __dict(self, origin, strict=True):
        """
        Process the dict merging.

        :param bool strict:
            Activates the strict mode.

        :rtype: dict
        """

        result = {}

        for index, data in self.main.items():
            if index in origin:
                if isinstance(data, dict) and isinstance(origin[index], dict):
                    result[index] = Merge(data).into(origin[index], strict=strict)
                elif isinstance(data, list) and isinstance(origin[index], list):
                    result[index] = Merge(data).into(origin[index], strict=strict)
                else:
                    result[index] = data
            else:
                result[index] = data

        for index, data in origin.items():
            if index not in result:
                result[index] = data

        return result

    def into(self, origin, strict=True):
        """
        Process the mergin.

        :param origin: The original data.
        :param bool strict:
            Activates the strict mode.
        """

        try:
            origin = origin.copy()
        except AttributeError:
            pass

        if isinstance(self.main, list) and isinstance(origin, list):
            return self.__list(origin, strict=strict)

        if isinstance(self.main, dict) and isinstance(origin, dict):
            return self.__dict(origin, strict=strict)

        return self.main  # pragma: no cover



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the percentage interface.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io//en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    MIT License

    Copyright (c) 2017, 2018, 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import PyFunceble


class Percentage:
    """
    Calculation of the percentage of each status.

    :param str domain_status: The status to increment.
    :param dict init:
        The data from a previous session we are continuing.

        .. warning::
            We expect the numbers and not the percentages.
    """

    def __init__(self, domain_status=None, init=None):
        # We get the status.
        self.status = domain_status

        if init and isinstance(init, dict):
            # * An information to init is given.
            # and
            # * It is a dictionnary.

            # We update the counter from the currently read data.
            PyFunceble.INTERN["counter"]["number"].update(init)

    def count(self):
        """
        Count the number of domain for each status.
        """

        if self.status:
            # The status is parsed.

            # We increase the number of tested.
            PyFunceble.INTERN["counter"]["number"]["tested"] += 1

            if (
                self.status.lower() in PyFunceble.STATUS.list.up
                or self.status.lower() in PyFunceble.STATUS.list.valid
                or self.status.lower() in PyFunceble.STATUS.list.sane
            ):
                # The status is in the list of up status.

                # We increase the number of up.
                PyFunceble.INTERN["counter"]["number"]["up"] += 1
            elif (
                self.status.lower() in PyFunceble.STATUS.list.down
                or self.status.lower() in PyFunceble.STATUS.list.malicious
            ):
                # The status is in the list of down status.

                # We increase the number of down.
                PyFunceble.INTERN["counter"]["number"]["down"] += 1
            else:
                # The status is not in the list of up nor down status.

                # We increase the number of invalid.
                PyFunceble.INTERN["counter"]["number"]["invalid"] += 1

    @classmethod
    def calculate(cls):
        """
        Calculate the percentage of each status.
        """

        PyFunceble.INTERN["counter"]["percentage"] = {}

        # We map the current state/counters of the different status.
        percentages = {
            x: PyFunceble.INTERN["counter"]["number"][x]
            for x in PyFunceble.STATUS.official.keys()
            if x in PyFunceble.INTERN["counter"]["number"]
        }

        for percentage in percentages:
            # We loop through our map index.

            # We calculate the percentage.
            calculation = (
                percentages[percentage]
                * 100
                // PyFunceble.INTERN["counter"]["number"]["tested"]
            )

            # And we update the percentage counter of the actual status.
            PyFunceble.INTERN["counter"]["percentage"].update({percentage: calculation})

        # raise Exception(PyFunceble.INTERN["counter"]["percentage"])

    def log(self):
        """
        Print on screen and on file the percentages for each status.
        """

        if (
            PyFunceble.CONFIGURATION.show_percentage
            and PyFunceble.INTERN["counter"]["number"]["tested"] > 0
        ):
            # * We are allowed to show the percentage on screen.
            # and
            # * The number of tested is greater than 0.

            # We initiate the output file.
            output = (
                PyFunceble.OUTPUT_DIRECTORY
                + PyFunceble.OUTPUTS.parent_directory
                + PyFunceble.OUTPUTS.logs.directories.parent
                + PyFunceble.OUTPUTS.logs.directories.percentage
                + PyFunceble.OUTPUTS.logs.filenames.percentage
            )

            # We delete the output file if it does exist.
            PyFunceble.helpers.File(output).delete()

            # We calculate the percentage of each statuses.
            self.calculate()

            # We construct the different lines/data to print on screen and file.
            lines_to_print = [
                [
                    PyFunceble.STATUS.official.up,
                    str(PyFunceble.INTERN["counter"]["percentage"]["up"]) + "%",
                    PyFunceble.INTERN["counter"]["number"]["up"],
                ],
                [
                    PyFunceble.STATUS.official.down,
                    str(PyFunceble.INTERN["counter"]["percentage"]["down"]) + "%",
                    PyFunceble.INTERN["counter"]["number"]["down"],
                ],
                [
                    PyFunceble.STATUS.official.invalid,
                    str(PyFunceble.INTERN["counter"]["percentage"]["invalid"]) + "%",
                    PyFunceble.INTERN["counter"]["number"]["invalid"],
                ],
            ]

            if PyFunceble.CONFIGURATION.syntax:
                # We are checking for syntax.

                # We update the denomination of the UP.
                lines_to_print[0][0] = PyFunceble.STATUS.official.valid
                lines_to_print[0][1] = (
                    str(PyFunceble.INTERN["counter"]["percentage"]["valid"]) + "%"
                )

                lines_to_print[0][2] = PyFunceble.INTERN["counter"]["number"]["valid"]

                # And we unset the INACTIVE line.
                del lines_to_print[1]
                del PyFunceble.INTERN["counter"]["number"]["down"]
                del PyFunceble.INTERN["counter"]["number"]["up"]

            if PyFunceble.CONFIGURATION.reputation:
                # We are checking for reputation.

                # We update the denomination of the UP.
                lines_to_print[0][0] = PyFunceble.STATUS.official.sane
                lines_to_print[0][1] = (
                    str(PyFunceble.INTERN["counter"]["percentage"]["sane"]) + "%"
                )

                lines_to_print[0][2] = PyFunceble.INTERN["counter"]["number"]["sane"]

                # We update the denomination of the Down.
                lines_to_print[1][0] = PyFunceble.STATUS.official.malicious
                lines_to_print[1][1] = (
                    str(PyFunceble.INTERN["counter"]["percentage"]["malicious"]) + "%"
                )
                lines_to_print[1][2] = PyFunceble.INTERN["counter"]["number"][
                    "malicious"
                ]

                # And we unset the INVALID line.
                del lines_to_print[2]
                del PyFunceble.INTERN["counter"]["number"]["invalid"]
                del PyFunceble.INTERN["counter"]["number"]["up"]
                del PyFunceble.INTERN["counter"]["number"]["down"]

            if (
                not PyFunceble.CONFIGURATION.quiet
                and not PyFunceble.CONFIGURATION.simple
            ):
                # * The quiet mode is not activated.
                # or
                # * We are testing in simple mode.

                # We print a new line.
                print("\n")

                # We print the percentage header on file and screen.
                PyFunceble.output.Prints(None, "Percentage", output).header()

                for to_print in lines_to_print:
                    # We loop throught the different line to print.
                    # (one line for each status.)

                    # And we print the current status line on file and screen.
                    PyFunceble.output.Prints(to_print, "Percentage", output).data()
            else:  # pragma: no cover
                # The quiet mode is activated.

                # We print the percentage header on file.
                PyFunceble.output.Prints(
                    None, "Percentage", output, only_on_file=True
                ).header()

                for to_print in lines_to_print:
                    # We loop throught the different line to print.
                    # (one line for each status.)

                    # And we print the current status line on file.
                    PyFunceble.output.Prints(
                        to_print, "Percentage", output, only_on_file=True
                    ).data()

        elif PyFunceble.INTERN["counter"]["number"]["tested"] > 0:
            # * We are not allowed to show the percentage on screen.
            # but
            # * The number of tested is greater than 0.

            # We run the calculation.
            # Note: The following is needed, because all counter calculation are
            # done by this class.
            self.calculate()



# pylint: disable=too-many-lines, line-too-long
"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the CLI entry points.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io//en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    MIT License

    Copyright (c) 2017, 2018, 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import argparse
import sys
from multiprocessing import set_start_method
from os import cpu_count

from colorama import Back, Fore, Style
from colorama import init as initiate_colorama

import PyFunceble

from .dispatcher import Dispatcher
from .iana import Iana
from .production import Production
from .publicsuffix import PublicSuffix


def tool():  # pragma: no cover pylint: disable=too-many-branches,too-many-statements,too-many-locals
    """
    Provides the CLI.
    """

    if __name__ == "PyFunceble.cli":
        if (
            PyFunceble.abstracts.Platform.is_windows()
            or PyFunceble.abstracts.Platform.is_mac_os()
        ):
            set_start_method("spawn")
        # We initiate the end of the coloration at the end of each line.
        initiate_colorama(autoreset=True)

        try:
            # The following handle the command line argument.

            try:
                # We load the configuration.
                PyFunceble.load_config(generate_directory_structure=False)

                preset = PyFunceble.cconfig.Preset()

                description = (
                    f"{Style.BRIGHT}{Fore.GREEN}PyFunceble{Style.RESET_ALL} - "
                    "The tool to check the availability or syntax of domains, IPv4, IPv6 or URL."
                )

                epilog = (
                    f"{Style.BRIGHT}{Fore.YELLOW}For an in-depth usage, explanation and examples of the arguments, "
                    f"you should read the documentation at{Fore.GREEN} https://pyfunceble.readthedocs.io//en/dev/"
                    f"{Style.RESET_ALL}\n\n"
                    f"Crafted with {Fore.RED}♥{Fore.RESET} by "
                    f"{Style.BRIGHT}{Fore.CYAN}Nissar Chababy (@funilrys){Style.RESET_ALL} "
                    f"with the help of{Style.BRIGHT}{Fore.GREEN} "
                    f"https://pyfunceble.github.io/contributors.html{Style.RESET_ALL} "
                    f"&& {Style.BRIGHT}{Fore.GREEN}"
                    "https://pyfunceble.github.io/special-thanks.html"
                )

                parser = argparse.ArgumentParser(
                    description=description,
                    epilog=epilog,
                    add_help=False,
                    formatter_class=argparse.RawTextHelpFormatter,
                )

                source_group = parser.add_argument_group("Source")
                filtering_group = parser.add_argument_group(
                    "Source filtering, decoding, conversion and expansion"
                )
                test_control = parser.add_argument_group("Test control")
                dns_control_group = parser.add_argument_group("DNS (resolver) control")
                database_control_group = parser.add_argument_group("Databases")
                output_control_group = parser.add_argument_group("Output control")
                multiprocessing_group = parser.add_argument_group("Multiprocessing")
                ci_group = parser.add_argument_group("CI / CD")
                unique_group = parser.add_argument_group("Unique actions")

                current_value_format = (
                    f"\n{Fore.YELLOW}{Style.BRIGHT}Configured value: {Fore.BLUE}"
                )

                source_group.add_argument(
                    "-d",
                    "--domain",
                    type=str,
                    nargs="+",
                    help="Test one or more domains, separated by spaces.",
                )

                source_group.add_argument(
                    "-u",
                    "--url",
                    type=str,
                    nargs="+",
                    help="Test one or more full URL, separated by spaces.",
                )

                source_group.add_argument(
                    "-f",
                    "--file",
                    type=str,
                    help="Read a local or remote (RAW link) file and test all domains inside it."
                    "\nIf remote (RAW link) file is given, PyFunceble will download it,\n "
                    "and test the content of the given RAW link as if it was a locally stored file.",
                )

                source_group.add_argument(
                    "-uf",
                    "--url-file",
                    type=str,
                    help="Read a local or remote (RAW link) file and test all (full) URLs inside it."
                    "\nIf remote (RAW link) file is given, PyFunceble will download it,\n "
                    "and test the content of the given RAW link as if it was a locally stored file. "
                    "\n\nThis argument test if an URL is available. It ONLY test full URLs.",
                )

                filtering_group.add_argument(
                    "-ad",
                    "--adblock",
                    action="store_true",
                    help="Switch the decoding of the adblock format. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.adblock)
                        + Style.RESET_ALL
                    ),
                )

                filtering_group.add_argument(
                    "--aggressive", action="store_true", help=argparse.SUPPRESS
                )

                filtering_group.add_argument(
                    "--complements",
                    action="store_true",
                    help="Switch the value of the generation and test of the complements. "
                    "\nA complement is for example `example.org` if `www.example.org` "
                    "is given and vice-versa. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.generate_complements)
                        + Style.RESET_ALL
                    ),
                )

                filtering_group.add_argument(
                    "--filter", type=str, help="Domain to filter (regex)."
                )

                filtering_group.add_argument(
                    "--idna",
                    action="store_true",
                    help="Switch the value of the IDNA conversion. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.idna_conversion)
                        + Style.RESET_ALL
                    ),
                )

                filtering_group.add_argument(
                    "--mining",
                    action="store_true",
                    help="Switch the value of the mining subsystem usage. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.mining)
                        + Style.RESET_ALL
                    ),
                )

                test_control.add_argument(
                    "" "-c",
                    "--auto-continue",
                    "--continue",
                    action="store_true",
                    help="Switch the value of the auto continue mode. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.auto_continue)
                        + Style.RESET_ALL
                    ),
                )

                test_control.add_argument(
                    "--cooldown-time",
                    type=float,
                    default=0,
                    help="Switch the value of the cooldown time to apply between each test. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.cooldown_time)
                        + Style.RESET_ALL
                    ),
                )

                test_control.add_argument(
                    "--http",
                    action="store_true",
                    help="Switch the value of the usage of HTTP code. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.http_codes.active)
                        + Style.RESET_ALL
                    ),
                )

                test_control.add_argument(
                    "--local",
                    action="store_true",
                    help="Switch the value of the local network testing. %s"
                    % (
                        current_value_format
                        + repr(preset.switch("local"))
                        + Style.RESET_ALL
                    ),
                )

                test_control.add_argument(
                    "-ns",
                    "--no-special",
                    action="store_true",
                    help="Switch the value of the usage of the SPECIAL rules. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.no_special)
                        + Style.RESET_ALL
                    ),
                )

                test_control.add_argument(
                    "-nw",
                    "--no-whois",
                    action="store_true",
                    help="Switch the value of the usage of WHOIS to test the domain's status. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.no_whois)
                        + Style.RESET_ALL
                    ),
                )

                test_control.add_argument(
                    "--syntax",
                    action="store_true",
                    help="Switch the value of the syntax test mode. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.syntax)
                        + Style.RESET_ALL
                    ),
                )

                test_control.add_argument(
                    "-t",
                    "--timeout",
                    type=float,
                    default=0,
                    help="Switch the value of the timeout. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.timeout)
                        + Style.RESET_ALL
                    ),
                )

                test_control.add_argument(
                    "--reputation",
                    action="store_true",
                    help="Switch the value of the reputation test mode. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.reputation)
                        + Style.RESET_ALL
                    ),
                )

                test_control.add_argument(
                    "--use-reputation-data",
                    action="store_true",
                    help="Switch the value of the reputation data usage. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.use_reputation_data)
                        + Style.RESET_ALL
                    ),
                )

                test_control.add_argument(
                    "-ua",
                    "--user-agent",
                    type=str,
                    help="Set the user-agent to use and set every time we "
                    "interact with everything which\nis not the logs sharing system.",
                )

                test_control.add_argument(
                    "-vsc",
                    "--verify-ssl-certificate",
                    action="store_true",
                    help="Switch the value of the verification of the "
                    "SSL/TLS certificate when testing for URL. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.verify_ssl_certificate)
                        + Style.RESET_ALL
                    ),
                )

                test_control.add_argument(
                    "--wildcard",
                    action="store_true",
                    help="Switch the value of the wildcards test.\n\n"
                    "When used, wildcards will be proprely tested.\n\n"
                    f"{Fore.MAGENTA}{Style.BRIGHT}Warning: This is not taken in consideration if the "
                    f"'--syntax' argument is not given.{Style.RESET_ALL} %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.wildcard)
                        + Style.RESET_ALL
                    ),
                )

                dns_control_group.add_argument(
                    "--dns",
                    nargs="+",
                    help="Set one or more DNS server(s) to use during testing. "
                    "Separated by spaces.\n\nTo specify a port number for the "
                    "DNS server you append it as :port [ip:port].\n\n"
                    "If no port is specified, the default DNS port (53) is used. %s"
                    % (
                        current_value_format
                        + repr(", ".join(PyFunceble.CONFIGURATION.dns_server))
                        if PyFunceble.CONFIGURATION.dns_server
                        else current_value_format
                        + "OS (declared) DNS server"
                        + Style.RESET_ALL
                    ),
                )

                dns_control_group.add_argument(
                    "--dns-lookup-over-tcp",
                    action="store_true",
                    help="Make all DNS queries with TCP. "
                    "%s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.dns_lookup_over_tcp)
                    ),
                )

                database_control_group.add_argument(
                    "-db",
                    "--database",
                    action="store_true",
                    help="Switch the value of the usage of a database to store "
                    "inactive domains of the currently tested list. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.inactive_database)
                        + Style.RESET_ALL
                    ),
                )

                database_control_group.add_argument(
                    "--database-type",
                    type=str,
                    help="Tell us the type of database to use. "
                    "\nYou can choose between the following: `json | mariadb | mysql` %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.db_type)
                        + Style.RESET_ALL
                    ),
                )

                database_control_group.add_argument(
                    "-dbr",
                    "--days-between-db-retest",
                    type=int,
                    help="Set the numbers of days between each retest of domains present "
                    "into inactive-db.json. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.days_between_db_retest)
                        + Style.RESET_ALL
                    ),
                )

                database_control_group.add_argument(
                    "-wdb",
                    "--whois-database",
                    action="store_true",
                    help="Switch the value of the usage of a database to store "
                    "whois data to avoid whois servers rate limit. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.whois_database)
                        + Style.RESET_ALL
                    ),
                )

                output_control_group.add_argument(
                    "-a",
                    "--all",
                    action="store_false",
                    help="Output all available information on the screen. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.less)
                        + Style.RESET_ALL
                    ),
                )

                output_control_group.add_argument(
                    "-ex",
                    "--execution",
                    action="store_true",
                    help="Switch the default value of the execution time showing. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.show_execution_time)
                        + Style.RESET_ALL
                    ),
                )

                output_control_group.add_argument(
                    "--hierarchical",
                    action="store_true",
                    help="Switch the value of the hierarchical sorting of the tested file. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.hierarchical_sorting)
                        + Style.RESET_ALL
                    ),
                )

                output_control_group.add_argument(
                    "-h",
                    "--host",
                    action="store_true",
                    help="Switch the value of the generation of hosts file. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.generate_hosts)
                        + Style.RESET_ALL
                    ),
                )

                output_control_group.add_argument(
                    "-ip",
                    type=str,
                    help="Change the IP to print in the hosts files with the given one. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.custom_ip)
                        + Style.RESET_ALL
                    ),
                )

                output_control_group.add_argument(
                    "--json",
                    action="store_true",
                    help="Switch the value of the generation "
                    "of the JSON formatted list of domains. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.generate_json)
                        + Style.RESET_ALL
                    ),
                )

                output_control_group.add_argument(
                    "--less",
                    action="store_true",
                    help="Output less informations on screen. %s"
                    % (
                        current_value_format
                        + repr(preset.switch("less"))
                        + Style.RESET_ALL
                    ),
                )

                output_control_group.add_argument(
                    "-nf",
                    "--no-files",
                    action="store_true",
                    help="Switch the value of the production of output files. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.no_files)
                        + Style.RESET_ALL
                    ),
                )

                output_control_group.add_argument(
                    "-nl",
                    "--no-logs",
                    action="store_true",
                    help="Switch the value of the production of logs files "
                    "in the case we encounter some errors. %s"
                    % (
                        current_value_format
                        + repr(not PyFunceble.CONFIGURATION.logs)
                        + Style.RESET_ALL
                    ),
                )

                output_control_group.add_argument(
                    "-nu",
                    "--no-unified",
                    action="store_true",
                    help="Switch the value of the production unified logs "
                    "under the output directory. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.unified)
                        + Style.RESET_ALL
                    ),
                )

                output_control_group.add_argument(
                    "--percentage",
                    action="store_true",
                    help="Switch the value of the percentage output mode. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.show_percentage)
                        + Style.RESET_ALL
                    ),
                )

                output_control_group.add_argument(
                    "--plain",
                    action="store_true",
                    help="Switch the value of the generation "
                    "of the plain list of domains. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.plain_list_domain)
                        + Style.RESET_ALL
                    ),
                )

                output_control_group.add_argument(
                    "--dots",
                    action="store_true",
                    help="Prints dots to stdout instead of giving the impression that we hang on. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.print_dots)
                        + Style.RESET_ALL
                    ),
                )

                output_control_group.add_argument(
                    "-q",
                    "--quiet",
                    action="store_true",
                    help="Run the script in quiet mode. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.quiet)
                        + Style.RESET_ALL
                    ),
                )

                output_control_group.add_argument(
                    "--share-logs",
                    action="store_true",
                    help="Switch the value of the sharing of logs. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.share_logs)
                        + Style.RESET_ALL
                    ),
                )

                output_control_group.add_argument(
                    "-s",
                    "--simple",
                    action="store_true",
                    help="Switch the value of the simple output mode. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.simple)
                        + Style.RESET_ALL
                    ),
                )

                output_control_group.add_argument(
                    "--split",
                    action="store_true",
                    help="Switch the value of the split of the generated output files. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.inactive_database)
                        + Style.RESET_ALL
                    ),
                )

                multiprocessing_group.add_argument(
                    "-m",
                    "--multiprocess",
                    action="store_true",
                    help="Switch the value of the usage of multiple processes. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.multiprocess)
                        + Style.RESET_ALL
                    ),
                )

                multiprocessing_group.add_argument(
                    "--multiprocess-merging-mode",
                    type=str,
                    help="Sets the multiprocess merging mode. "
                    "\nYou can choose between the following: `live|ends`. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.multiprocess_merging_mode)
                        + Style.RESET_ALL
                    ),
                )

                multiprocessing_group.add_argument(
                    "-p",
                    "--processes",
                    type=int,
                    help="Set the number of simultaneous processes to use while "
                    "using multiple processes.\nIf omited, the number of "
                    "available CPU cores will be used instead.%s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.maximal_processes)
                        + Style.RESET_ALL
                    ),
                )

                ci_group.add_argument(
                    "--autosave-minutes",
                    type=int,
                    help="Update the minimum of minutes before we start "
                    "committing to upstream under the CI mode. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.ci_autosave_minutes)
                        + Style.RESET_ALL
                    ),
                )

                ci_group.add_argument(
                    "--travis", action="store_true", help=argparse.SUPPRESS,
                )

                ci_group.add_argument(
                    "--ci",
                    action="store_true",
                    help="Switch the value of the CI mode. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.ci)
                        + Style.RESET_ALL
                    ),
                )

                ci_group.add_argument(
                    "--ci-branch",
                    type=str,
                    default="master",
                    help="Switch the branch name where we are going to push. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.ci_branch)
                        + Style.RESET_ALL
                    ),
                )

                ci_group.add_argument(
                    "--travis-branch",
                    type=str,
                    default="master",
                    help=argparse.SUPPRESS,
                )

                ci_group.add_argument(
                    "--ci-distribution-branch",
                    type=str,
                    default="master",
                    help="Switch the branch name where we are going to push the final results. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.ci_distribution_branch)
                        + Style.RESET_ALL
                    ),
                )

                ci_group.add_argument(
                    "--travis-distribution-branch",
                    type=str,
                    default="master",
                    help=argparse.SUPPRESS,
                )

                ci_group.add_argument(
                    "--cmd",
                    type=str,
                    help="Pass a command to run before each commit "
                    "(except the final one) under the CI mode. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.command_before_end)
                        + Style.RESET_ALL
                    ),
                )

                ci_group.add_argument(
                    "--cmd-before-end",
                    type=str,
                    help="Pass a command to run before the results "
                    "(final) commit under the CI mode. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.command_before_end)
                        + Style.RESET_ALL
                    ),
                )

                ci_group.add_argument(
                    "--commit-autosave-message",
                    type=str,
                    help="Replace the default autosave commit message. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.travis_autosave_commit)
                        + Style.RESET_ALL
                    ),
                )

                ci_group.add_argument(
                    "--commit-results-message",
                    type=str,
                    help="Replace the default results (final) commit message. %s"
                    % (
                        current_value_format
                        + repr(PyFunceble.CONFIGURATION.travis_autosave_final_commit)
                        + Style.RESET_ALL
                    ),
                )

                unique_group.add_argument(
                    "--clean",
                    action="store_true",
                    help="Clean all files under the output directory.",
                )

                unique_group.add_argument(
                    "--clean-all",
                    action="store_true",
                    help="Clean all files under the output directory "
                    "along with all file generated by PyFunceble.",
                )

                unique_group.add_argument(
                    "--directory-structure",
                    action="store_true",
                    help="Generate the directory and files that are needed and which does "
                    "not exist in the current directory.",
                )

                unique_group.add_argument(
                    "--iana", action="store_true", help=argparse.SUPPRESS,
                )

                unique_group.add_argument(
                    "--production", action="store_true", help=argparse.SUPPRESS
                )

                unique_group.add_argument(
                    "-psl",
                    "--public-suffix",
                    action="store_true",
                    help=argparse.SUPPRESS,
                )

                unique_group.add_argument(
                    "--generate-files-from-database",
                    action="store_true",
                    help=argparse.SUPPRESS,
                )

                unique_group.add_argument(
                    "--generate-all-files-from-database",
                    action="store_true",
                    help=argparse.SUPPRESS,
                )

                parser.add_argument(
                    "--debug", action="store_true", help=argparse.SUPPRESS
                )

                parser.add_argument(
                    "--help",
                    action="help",
                    default=argparse.SUPPRESS,
                    help="Show this help message and exit.",
                )

                parser.add_argument(
                    "-v",
                    "--version",
                    help="Show the version of PyFunceble and exit.",
                    action="version",
                    version="%(prog)s " + PyFunceble.VERSION,
                )

                args = parser.parse_args()

                if args.debug:
                    PyFunceble.CONFIGURATION.debug = preset.switch("debug")
                    PyFunceble.LOGGER.authorized = PyFunceble.LOGGER.authorization(None)
                    PyFunceble.LOGGER.init()

                if args.less:
                    PyFunceble.CONFIGURATION.less = args.less
                elif not args.all:
                    PyFunceble.CONFIGURATION.less = args.all

                if args.adblock:
                    PyFunceble.CONFIGURATION.adblock = preset.switch("adblock")

                if args.aggressive:
                    PyFunceble.CONFIGURATION.aggressive = preset.switch("aggressive")

                if args.auto_continue:
                    PyFunceble.CONFIGURATION.auto_continue = preset.switch(
                        "auto_continue"
                    )

                if args.autosave_minutes:
                    PyFunceble.CONFIGURATION.ci_autosave_minutes = args.autosave_minutes

                if args.cmd:
                    PyFunceble.CONFIGURATION.command = args.cmd

                if args.cmd_before_end:
                    PyFunceble.CONFIGURATION.command_before_end = args.cmd_before_end

                if args.commit_autosave_message:
                    PyFunceble.CONFIGURATION.ci_autosave_commit = (
                        args.commit_autosave_message
                    )

                if args.commit_results_message:
                    PyFunceble.CONFIGURATION.ci_autosave_final_commit = (
                        args.commit_results_message
                    )

                if args.complements:
                    PyFunceble.CONFIGURATION.generate_complements = preset.switch(
                        "generate_complements"
                    )

                if args.cooldown_time >= 0:
                    PyFunceble.CONFIGURATION.cooldown_time = args.cooldown_time

                if args.database:
                    PyFunceble.CONFIGURATION.inactive_database = preset.switch(
                        "inactive_database"
                    )

                if args.database_type:
                    if args.database_type.lower() in ["json", "mariadb", "mysql"]:
                        PyFunceble.CONFIGURATION.db_type = args.database_type.lower()
                    else:
                        print(
                            Style.BRIGHT
                            + Fore.RED
                            + "Unknown database type: {0}".format(
                                repr(args.database_type)
                            )
                        )
                        sys.exit(1)

                if args.days_between_db_retest:
                    PyFunceble.CONFIGURATION.days_between_db_retest = (
                        args.days_between_db_retest
                    )

                if args.dns:
                    PyFunceble.CONFIGURATION.dns_server = args.dns

                if args.dns_lookup_over_tcp:
                    PyFunceble.CONFIGURATION.dns_lookup_over_tcp = preset.switch(
                        "dns_lookup_over_tcp"
                    )

                if args.execution:
                    PyFunceble.CONFIGURATION.show_execution_time = preset.switch(
                        "show_execution_time"
                    )

                if args.filter:
                    PyFunceble.CONFIGURATION.filter = args.filter

                if args.hierarchical:
                    PyFunceble.CONFIGURATION.hierarchical_sorting = preset.switch(
                        "hierarchical_sorting"
                    )

                if args.host:
                    PyFunceble.CONFIGURATION.generate_hosts = preset.switch(
                        "generate_hosts"
                    )

                if args.http:
                    PyFunceble.CONFIGURATION.http_codes.active = preset.switch(
                        PyFunceble.CONFIGURATION.http_codes.active, True
                    )

                if args.idna:
                    PyFunceble.CONFIGURATION.idna_conversion = preset.switch(
                        "idna_conversion"
                    )

                if args.ip:
                    PyFunceble.CONFIGURATION.custom_ip = args.ip

                if args.json:
                    PyFunceble.CONFIGURATION.generate_json = preset.switch(
                        "generate_json"
                    )

                if args.local:
                    PyFunceble.CONFIGURATION.local = preset.switch("local")

                if args.mining:
                    PyFunceble.CONFIGURATION.mining = preset.switch("mining")

                if args.multiprocess:
                    PyFunceble.CONFIGURATION.multiprocess = preset.switch(
                        "multiprocess"
                    )

                if args.multiprocess_merging_mode:
                    if args.multiprocess_merging_mode.lower() in ["end", "live"]:
                        PyFunceble.CONFIGURATION.multiprocess_merging_mode = (
                            args.multiprocess_merging_mode.lower()
                        )
                    else:
                        print(
                            Style.BRIGHT
                            + Fore.RED
                            + "Unknown multiprocess merging mode: {0}".format(
                                repr(args.multiprocess_merging_mode)
                            )
                        )
                        sys.exit(1)

                if args.no_files:
                    PyFunceble.CONFIGURATION.no_files = preset.switch("no_files")

                if args.no_logs:
                    PyFunceble.CONFIGURATION.logs = preset.switch("logs")

                if args.no_special:
                    PyFunceble.CONFIGURATION.no_special = preset.switch("no_special")

                if args.no_unified:
                    PyFunceble.CONFIGURATION.unified = preset.switch("unified")

                if args.no_whois:
                    PyFunceble.CONFIGURATION.no_whois = preset.switch("no_whois")

                if args.percentage:
                    PyFunceble.CONFIGURATION.show_percentage = preset.switch(
                        "show_percentage"
                    )

                if args.plain:
                    PyFunceble.CONFIGURATION.plain_list_domain = preset.switch(
                        "plain_list_domain"
                    )

                if args.dots:
                    PyFunceble.CONFIGURATION.print_dots = preset.switch("print_dots")

                if args.processes:
                    PyFunceble.CONFIGURATION.maximal_processes = args.processes
                else:
                    PyFunceble.CONFIGURATION.maximal_processes = cpu_count()

                if args.quiet:
                    PyFunceble.CONFIGURATION.quiet = preset.switch("quiet")

                if args.reputation:
                    PyFunceble.CONFIGURATION.reputation = preset.switch("reputation")

                if args.share_logs:
                    PyFunceble.CONFIGURATION.share_logs = preset.switch("share_logs")

                if args.simple:
                    PyFunceble.CONFIGURATION.simple = preset.switch("simple")

                if args.split:
                    PyFunceble.CONFIGURATION.split = preset.switch("split")

                if args.syntax:
                    PyFunceble.CONFIGURATION.syntax = preset.switch("syntax")

                if args.timeout >= 0:
                    PyFunceble.CONFIGURATION.timeout = args.timeout

                if args.use_reputation_data:
                    PyFunceble.CONFIGURATION.use_reputation_data = preset.switch(
                        "use_reputation_data"
                    )

                if args.ci or args.travis:
                    PyFunceble.CONFIGURATION.ci = preset.switch("ci")

                if args.ci_distribution_branch:
                    PyFunceble.CONFIGURATION.ci_distribution_branch = (
                        args.ci_distribution_branch
                    )
                elif args.travis_distribution_branch:
                    PyFunceble.CONFIGURATION.ci_distribution_branch = (
                        args.travis_distribution_branch
                    )

                if args.ci_branch:
                    PyFunceble.CONFIGURATION.ci_branch = args.ci_branch
                elif args.travis_branch:
                    PyFunceble.CONFIGURATION.ci_branch = args.travis_branch

                if args.user_agent:
                    PyFunceble.CONFIGURATION.user_agent.custom = args.user_agent

                if args.verify_ssl_certificate:
                    PyFunceble.CONFIGURATION.verify_ssl_certificate = (
                        args.verify_ssl_certificate
                    )

                if args.whois_database:
                    PyFunceble.CONFIGURATION.whois_database = preset.switch(
                        "whois_database"
                    )

                if args.wildcard:
                    PyFunceble.CONFIGURATION.wildcard = preset.switch("wildcard")

                PyFunceble.core.CLI.colorify_logo(home=True)

                preset.timeout()
                preset.cooldown_time()
                preset.dns_lookup_over_tcp()

                if args.clean:
                    PyFunceble.output.Clean()

                if args.clean_all:
                    PyFunceble.output.Clean(args.clean_all)

                if args.directory_structure:
                    PyFunceble.output.Constructor()

                if args.iana:
                    Iana().update()

                if args.production:
                    Production()

                if args.public_suffix:
                    PublicSuffix().update()

                PyFunceble.LOGGER.info(f"ARGS:\n{args}")

                # We compare the versions (upstream and local) and in between.
                PyFunceble.core.CLI.compare_version_and_print_messages()

                # We call our Core which will handle all case depending of the configuration or
                # the used command line arguments.
                Dispatcher(
                    preset,
                    domain_or_ip=args.domain,
                    file_path=args.file,
                    url_to_test=args.url,
                    url_file_path=args.url_file,
                    generate_results_only=args.generate_files_from_database,
                    generate_all_results_only=args.generate_all_files_from_database,
                )
            except Exception as exception:
                try:
                    PyFunceble.LOGGER.exception()
                except AttributeError:
                    pass

                raise exception

        except KeyboardInterrupt:
            PyFunceble.core.CLI.stay_safe()



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the command helpers. This helpers let us run Shell commands.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io//en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    MIT License

    Copyright (c) 2017, 2018, 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import sys
from os import environ
from subprocess import PIPE, STDOUT, Popen


class Command:
    """
    Shell command execution.

    :param str command: The command to execute.
    :param str encoding: The encoding to use to decode the shell output.
    """

    def __init__(self, command, encoding="utf-8"):  # pragma: no cover
        # We set the default decoding type.
        self.decode_type = encoding

        if isinstance(command, list):
            # The given command is a list.

            # We construct the command we are going to run.
            self.command = " ".join(command)
        elif isinstance(command, str):
            # The given command is a string.

            # We set the command we are going to run.
            self.command = command.strip()
        else:
            raise NotImplementedError(
                "Unknown command type: `{}`".format(type(command))
            )

    def _decode_output(self, to_decode):
        """
        Decode the output of a shell command in order to be readable.

        :param bytes to_decode: Output of a command to decode.

        :return: The decoded output.
        :rtype: str
        """

        if to_decode:
            return to_decode.decode(self.decode_type)

        return ""

    def execute(self):
        """
        Execute the given command.

        :return: The output of the command.
        :rtype: str
        """

        # We initiate a process and parse the command to it.
        with Popen(
            self.command, stdout=PIPE, stderr=STDOUT, shell=True, env=environ
        ) as process:
            # We communicate the command and get the output and the error.
            (stdout, stderr) = process.communicate()

            if process.returncode != 0:  # pragma: no cover
                # The return code is different to 0.

                # We return the decoded error.
                return self._decode_output(stderr)

            # The return code (or exit code if you prefer) if equal to 0.

            # We return the decoded output of the executed command.
            return self._decode_output(stdout)

    def run(self, rstrip=True):
        """
        Run the given command and yield each line(s) one by one.

        .. note::
            The difference between this method and :func:`~PyFunceble.helpers.Command.execute`
            is that :func:`~PyFunceble.helpers.Command.execute` wait for the process to end
            in order to return its output while this method return each line one by one
            - as they are outputed.

        :param bool rstrip:
            Deactivates the rstrip of the output.
        """

        with Popen(
            self.command, stdout=PIPE, stderr=STDOUT, shell=True, env=environ
        ) as process:
            # We initiate a process and parse the command to it.

            while True:
                # We loop infinitly because we want to get the output
                # until there is none.

                # We get the current line from the process stdout.
                #
                # Note: we use rstrip() because we are paranoid :-)
                current_line = process.stdout.readline()

                if not current_line and process.poll() is not None:
                    # The current line is empty or equal to None.

                    # We break the loop.
                    break

                # The line is not empty nor equal to None.

                if rstrip:
                    # We encode and yield the current line
                    yield self._decode_output(current_line.rstrip())
                else:
                    # We encode and yield the current line
                    yield self._decode_output(current_line)

    def run_to_stdout(self):
        """
        Run the given command and print each line(s) to stdout.
        """

        for line in self.run(rstrip=False):
            sys.stdout.write(line)



# pylint:disable=line-too-long
"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Tests of PyFunceble.abstracts.package

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io//en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    MIT License

    Copyright (c) 2017, 2018, 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

# pylint: enable=line-too-long

from unittest import TestCase
from unittest import main as launch_tests
from unittest.mock import patch

from PyFunceble.abstracts import Version


class TestVersion(TestCase):
    """
    Tests of PyFunceble.abstracts.Version
    """

    def test_split_version(self):
        """
        Tests the case that we want to split the version.
        """

        given = "1.0.0.dev (Hello, World!)"
        expected = ["1", "0", "0"]
        actual = Version.split_versions(given)

        self.assertEqual(expected, actual)

    def test_split_version_with_non_digits(self):
        """
        Tests the case that we want to split the version
        but also have the code name.
        """

        given = "1.0.0.dev (Hello, World!)"
        expected = (["1", "0", "0"], "dev (Hello, World!)")
        actual = Version.split_versions(given, return_non_digits=True)

        self.assertEqual(expected, actual)

    def test_literal_comparison(self):
        """
        Tests the literal comparison.
        """

        given = "1.0.0.dev (Hello, World!)"
        expected = True
        actual = Version.literally_compare(given, given)

        self.assertEqual(expected, actual)

    def test_literal_comparison_different(self):
        """
        Tests the litaral comparison for the case that both given version are different.
        """

        given = "1.0.0.dev (Hello, World!)"
        expected = False
        actual = Version.literally_compare(given, given.replace(".", "_"))

        self.assertEqual(expected, actual)

    @patch("PyFunceble.abstracts.Package.VERSION", "1.0.0.dev (Hello, World)")
    def test_compare_local_version_is_same(self):
        """
        Tests the comparison for the case that the local version is older.
        """

        given = "1.0.0.dev (Hello, World)"
        expected = None
        actual = Version.compare(given)

        self.assertEqual(expected, actual)

    @patch("PyFunceble.abstracts.Package.VERSION", "1.50.0.dev (Hello, World)")
    def test_compare_local_version_is_older(self):
        """
        Tests the comparison for the case that the local version is older.
        """

        given = "2.34.0.dev (Hello, World)"
        expected = True
        actual = Version.compare(given)

        self.assertEqual(expected, actual)

    @patch("PyFunceble.abstracts.Package.VERSION", "2.10.0.dev (Hello, World)")
    def test_compare_local_version_is_newer(self):
        """
        Tests the comparison for the case that the local version is older.
        """

        given = "1.15.0.dev (Hello, World)"
        expected = False
        actual = Version.compare(given)

        self.assertEqual(expected, actual)

    @patch("PyFunceble.abstracts.Package.VERSION", "2.10.0.dev (Hello, World)")
    def test_is_local_dev(self):
        """
        Tests if the local version is the dev one.
        """

        expected = True
        actual = Version.is_local_dev()

        self.assertEqual(expected, actual)

    @patch("PyFunceble.abstracts.Package.VERSION", "2.10.0. (Hello, World)")
    def test_is_not_local_dev(self):
        """
        Tests if the local version is the not the dev one.
        """

        expected = False
        actual = Version.is_local_dev()

        self.assertEqual(expected, actual)


if __name__ == "__main__":
    launch_tests()



# pylint:disable=line-too-long
"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Tests of PyFunceble.database.inactive

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io//en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    MIT License

    Copyright (c) 2017, 2018, 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""
# pylint: enable=line-too-long

from datetime import datetime, timedelta
from time import sleep
from unittest import TestCase
from unittest import main as launch_tests

import PyFunceble
from PyFunceble.database.inactive import InactiveDB


class TestInactiveDB(TestCase):
    """
    Tests of PyFunceble.database.inactive
    """

    def setUp(self):
        """
        Setups everything needed for the tests.
        """

        PyFunceble.load_config(
            generate_directory_structure=False, custom={"db_type": "json"}
        )

        self.file_to_test = "this_is_a_ghost"
        self.storage_file = (
            PyFunceble.CONFIG_DIRECTORY + PyFunceble.OUTPUTS.default_files.inactive_db
        )

        self.time_past = str(int((datetime.now() - timedelta(days=365)).timestamp()))
        self.time_future = str(int((datetime.now() + timedelta(days=365)).timestamp()))

        self.inactive_db = InactiveDB(self.file_to_test, parent_process=True)

        PyFunceble.helpers.File(self.storage_file).delete()

    def tearDown(self):
        """
        Setups everything needed after a test.
        """

        PyFunceble.helpers.File(self.storage_file).delete()

    def test_load_file_does_not_exists(self):
        """
        Tests the case that we load the file but it does not exists.
        """

        self.inactive_db.load()
        expected = {self.file_to_test: {}}

        self.assertEqual(expected, self.inactive_db.database)

    def test_load_file_exists(self):
        """
        Tests the case that we load the file.
        """

        # We also test the merging.
        self.inactive_db.database = {
            self.file_to_test: {
                "to_test": {"google.com": PyFunceble.STATUS.official.invalid}
            },
            "this_is_another_ghost": {
                "190": {"||google.com^": PyFunceble.STATUS.official.invalid}
            },
        }

        expected = {
            self.file_to_test: {
                "0": {
                    "mÿethèrwallét.com": PyFunceble.STATUS.official.invalid,
                    "||google.com^": PyFunceble.STATUS.official.invalid,
                }
            },
            "this_is_another_ghost": {
                "0": {"||google.com^": PyFunceble.STATUS.official.invalid},
                "190": {"||google.com^": PyFunceble.STATUS.official.invalid},
            },
            "this_is_a_hidden_ghost": {
                "200": {"200.com": PyFunceble.STATUS.official.invalid}
            },
        }

        PyFunceble.helpers.Dict(expected.copy()).to_json_file(self.storage_file)
        expected[self.file_to_test][
            str(
                int(
                    (
                        datetime.now() - timedelta(days=1) - timedelta(seconds=100)
                    ).timestamp()
                )
            )
        ] = {"google.com": PyFunceble.STATUS.official.invalid}

        self.inactive_db.load()

        self.assertEqual(expected, self.inactive_db.database)

    def test_save(self):
        """
        Tests teh case that we save the database.
        """

        expected = {
            self.file_to_test: {
                "0": {
                    "mÿethèrwallét.com": PyFunceble.STATUS.official.invalid,
                    "||google.com^": PyFunceble.STATUS.official.invalid,
                }
            }
        }

        self.inactive_db.database = expected.copy()
        self.inactive_db.save()

        self.assertEqual(
            expected, PyFunceble.helpers.Dict().from_json_file(self.storage_file)
        )

    def test_initiate_tested_path_does_not_exists(self):
        """
        Tests the initiate method for the case that the file
        we are testing is not indexed.
        """

        expected = {self.file_to_test: {}}

        self.inactive_db.database = {}
        self.inactive_db.initiate()

        self.assertEqual(expected, self.inactive_db.database)

    def test_add_tested_path_does_not_exists(self):
        """
        Tests the addition method for the case that teh flle
        we are testing is not indexed.
        """

        self.inactive_db.database = {}

        subject = "hello.world"

        expected = {
            self.file_to_test: {
                str(int(datetime.now().timestamp())): {
                    subject: PyFunceble.STATUS.official.down
                }
            }
        }

        self.inactive_db.add(subject, PyFunceble.STATUS.official.down)
        self.assertEqual(expected, self.inactive_db.database)

        subject = "world.hello"

        sleep(2)

        expected[self.file_to_test][str(int(datetime.now().timestamp()))] = {
            subject: PyFunceble.STATUS.official.down
        }

        self.inactive_db.add(subject, PyFunceble.STATUS.official.down)
        self.assertEqual(expected, self.inactive_db.database)

    def test_remove(self):
        """
        Tests of the deletion method.
        """

        subject = "hello.world"

        self.inactive_db.database = {
            self.file_to_test: {
                str(int(datetime.now().timestamp())): {
                    subject: PyFunceble.STATUS.official.down
                }
            }
        }

        expected = {self.file_to_test: {}}

        self.inactive_db.remove(subject)
        self.assertEqual(expected, self.inactive_db.database)

    def test_is_present(self):
        """
        Tests the presence checker method.
        """

        subject = "hello.world"

        self.inactive_db.database = {
            self.file_to_test: {
                str(int(datetime.now().timestamp())): {
                    subject: PyFunceble.STATUS.official.down
                }
            }
        }

        expected = True
        actual = subject in self.inactive_db

        self.assertEqual(expected, actual)

        expected = False
        actual = "world.hello" in self.inactive_db

        self.assertEqual(expected, actual)

    def test_set_item_from_outside(self):
        """
        Tests the case that we adds items from outside the class.
        """

        expected = {
            self.file_to_test: {
                str(int(datetime.now().timestamp())): {
                    "hello.world": PyFunceble.STATUS.official.down,
                    "world.hello": PyFunceble.STATUS.official.down,
                }
            }
        }

        self.inactive_db.database = {
            self.file_to_test: {
                str(int(datetime.now().timestamp())): {
                    "world.hello": PyFunceble.STATUS.official.down
                }
            }
        }

        self.inactive_db[str(int(datetime.now().timestamp()))] = {
            "hello.world": PyFunceble.STATUS.official.down
        }

        self.assertEqual(expected, self.inactive_db.database)


if __name__ == "__main__":
    launch_tests()



# pylint:disable=invalid-name, cyclic-import
"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io//en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    MIT License

    Copyright (c) 2017, 2018, 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import warnings
from datetime import datetime
from os import path
from os import sep as directory_separator
from time import sleep

from colorama import Fore
from dotenv import load_dotenv

import PyFunceble.abstracts as abstracts
import PyFunceble.config as cconfig
import PyFunceble.converter as converter
import PyFunceble.core as core
import PyFunceble.database as database
import PyFunceble.downloader as downloader
import PyFunceble.engine as engine
import PyFunceble.exceptions as exceptions
import PyFunceble.extractor as extractor
import PyFunceble.helpers as helpers
import PyFunceble.lookup as lookup
import PyFunceble.output as output
import PyFunceble.status as status
from PyFunceble.check import Check

# We set our project name.
NAME = abstracts.Package.NAME
# We set out project version.
VERSION = abstracts.Package.VERSION


if helpers.EnvironmentVariable("PYFUNCEBLE_CONFIG_DIR").exists():  # pragma: no cover
    # We handle the case that the `PYFUNCEBLE_CONFIG_DIR` environnement variable is set.
    CONFIG_DIRECTORY = helpers.EnvironmentVariable("PYFUNCEBLE_CONFIG_DIR").get_value()
elif helpers.EnvironmentVariable("PYFUNCEBLE_OUTPUT_DIR").exists():  # pragma: no cover
    # We hande the retro compatibility.
    CONFIG_DIRECTORY = helpers.EnvironmentVariable("PYFUNCEBLE_OUTPUT_DIR").get_value()
elif abstracts.Version.is_local_cloned():  # pragma: no cover
    # We handle the case that we are in a cloned.
    CONFIG_DIRECTORY = helpers.Directory.get_current(with_end_sep=True)
elif helpers.EnvironmentVariable("TRAVIS_BUILD_DIR").exists():  # pragma: no cover
    # We handle the case that we are under Travis CI.
    CONFIG_DIRECTORY = helpers.Directory.get_current(with_end_sep=True)
elif (
    helpers.EnvironmentVariable("CI_PROJECT_DIR").exists()
    and helpers.EnvironmentVariable("GITLAB_CI").exists()
):  # pragma: no cover
    # We handle the case that we are under GitLab CI/CD.
    CONFIG_DIRECTORY = helpers.Directory.get_current(with_end_sep=True)
else:  # pragma: no cover
    # We handle all other case and distributions specific cases.

    if abstracts.Platform.is_unix():
        # We are under a Linux distribution.

        # We set the default configuration location path.
        config_dir_path = (
            path.expanduser("~" + directory_separator + ".config") + directory_separator
        )

        if helpers.Directory(config_dir_path).exists():
            # Everything went right:
            #   * `~/.config` exists.
            # We set our configuration location path as the directory we are working with.
            CONFIG_DIRECTORY = config_dir_path
        elif helpers.Directory(path.expanduser("~")).exists():
            # Something went wrong:
            #   * `~/.config` does not exists.
            #   * `~` exists.
            # We set `~/` as the directory we are working with.
            #
            # Note: The `.` at the end is because we want to hide the directory we are
            # going to create.
            CONFIG_DIRECTORY = (
                path.expanduser("~") + directory_separator + "."
            )  # pylint: disable=line-too-long
        else:
            # Everything went wrong:
            #   * `~/.config` does not exists.
            #   * `~` soes not exists.
            # We set the current directory as the directory we are working with.
            CONFIG_DIRECTORY = helpers.Directory.get_current(with_end_sep=True)
    elif abstracts.Platform.is_windows():
        # We are under Windows or CygWin.

        if helpers.EnvironmentVariable("APPDATA").exists():
            # Everything went right:
            #   * `APPDATA` is into the environnement variables.
            # We set it as the directory we are working with.
            CONFIG_DIRECTORY = helpers.EnvironmentVariable("APPDATA").get_value()
        else:
            # Everything went wrong:
            #   * `APPDATA` is not into the environnement variables.
            # We set the current directory as the directory we are working with.
            CONFIG_DIRECTORY = helpers.Directory.get_current(with_end_sep=True)

    if not CONFIG_DIRECTORY.endswith(directory_separator):
        # If the directory we are working with does not ends with the directory
        # separator, we append it to the end.
        CONFIG_DIRECTORY += directory_separator

    # We append the name of the project to the directory we are working with.
    CONFIG_DIRECTORY += NAME + directory_separator

    if not helpers.Directory(CONFIG_DIRECTORY).exists():
        # If the directory does not exist we create it.
        helpers.Directory(CONFIG_DIRECTORY).create()

if not CONFIG_DIRECTORY.endswith(directory_separator):  # pragma: no cover
    # Again for safety, if the directory we are working with does not ends with
    # the directory separator, we append it to the end.
    CONFIG_DIRECTORY += directory_separator

# We set the location of the `output` directory which should always be in the current
# directory.
OUTPUT_DIRECTORY = helpers.Directory.get_current(with_end_sep=True)

# We initiate the location where we are going to save our whole configuration content.
CONFIGURATION = None
# We initiate the location where we are going to get all statuses.
STATUS = None
# We initiate the location where we are going to get all outputs.
OUTPUTS = None
# We initiate the location where we are going to get the map of the classification
# of each status codes for the analytic part.
HTTP_CODE = None
# We initiate the location where we are going to get all links.
LINKS = None
# We initiate a location which will have all internal data.
INTERN = None
# We initiate the location of the Logger.
LOGGER = None
# We initiate the location of the HTTP requests.
REQUESTS = None
# We initiate the DNS resolver.
DNSLOOKUP = None
# We initiate the PSL lookup.
PSLOOOKUP = None
# We initiate the IANA lookup.
IANALOOKUP = None
# We initate the loader.
LOADER = None

load_dotenv()
load_dotenv(CONFIG_DIRECTORY + ".env")
load_dotenv(CONFIG_DIRECTORY + abstracts.Infrastructure.ENV_FILENAME)

# We initiate the CLI logo of PyFunceble.
ASCII_PYFUNCEBLE = """
██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝
"""


def test(subject, complete=False, config=None):  # pragma: no cover
    """
    Test the availability of the given subject (domain or IP).

    :param subject: The subject (IP or domain) to test.
    :type subject: str|list

    :param bool complete:
        Activate the return of a dict with some significant data from
        the test.

    :param dict config:
        A dict with the configuration index (from .PyFunceble.yaml) to update.

    :return: The status or the informations of the domain.
    :rtype: str|dict

    .. note::
        If :code:`config` is given, the given :code:`dict` overwrite
        the last value of the given indexes in the configuration.

        It's actually something like following:

        ::

            pyfunceble.configuration.update(config_given_by_user)

    .. note::
        If :code:`complete` is set to :code:`True`, we return the following indexes.

        ::

            {
                "_status_source": None,
                "_status": None,
                "dns_lookup": [],
                "domain_syntax_validation": None,
                "expiration_date": None,
                "http_status_code": None,
                "ipv4_range_syntax_validation": None,
                "ipv4_syntax_validation": None,
                "ipv6_range_syntax_validation": None,
                "ipv6_syntax_validation": None,
                "status": None,
                "status_source": None,
                "subdomain_syntax_validation": None,
                "tested": None,
                "url_syntax_validation": None,
                "whois_record": None,
                "whois_server": None,
            }
    """

    if subject:
        # The subject is not empty nor None.

        # We return the status of the given subject.
        return core.API(
            subject, complete=complete, configuration=config
        ).domain_and_ip()

    # We return None, there is nothing to test.
    return None


def url_test(subject, complete=False, config=None):  # pragma: no covere
    """
    Test the availability of the given subject (URL).

    :param subject: The subject (URL) to test.
    :type subject: str|list

    :param bool complete:
        Activate the return of a dict with some significant data from
        the test.

    :param dict config:
        A dict with the configuration index (from .PyFunceble.yaml) to update.

    :return: The status or the informations of the URL.
    :rtype: str|dict

    .. note::
        If :code:`config` is given, the given :code:`dict` overwrite
        the last value of the given indexes in the configuration.

        It's actually something like following:

        ::

            pyfunceble.configuration.update(config_given_by_user)

    .. note::
        If :code:`complete` is set to :code:`True`, we return the following indexes.

        ::

            {
                "_status_source": None,
                "_status": None,
                "dns_lookup": [],
                "domain_syntax_validation": None,
                "expiration_date": None,
                "http_status_code": None,
                "ipv4_range_syntax_validation": None,
                "ipv4_syntax_validation": None,
                "ipv6_range_syntax_validation": None,
                "ipv6_syntax_validation": None,
                "status": None,
                "status_source": None,
                "subdomain_syntax_validation": None,
                "tested": None,
                "url_syntax_validation": None,
                "whois_record": None,
                "whois_server": None,
            }
    """

    if subject:
        # The given URL is not empty nor None.

        # We retunr the status of the the url.
        return core.API(subject, complete=complete, configuration=config).url()

    # We return None, there is nothing to test.
    return None


def dns_lookup(
    subject, dns_server=None, complete=False, lifetime=3
):  # pragma: no cover
    """
    Make a DNS lookup of the given subject.

    :param str subject: The subject we are working with.
    :param dns_server: A (or list of) DNS server to use while resolving.
    :type dns_server: str|int
    :param bool complete:
        Tell us to look for everything instead of :code:`NS` only.
    :param int lifetime: The query lifetime.

    :return:
        A dict with following index if the given subject is not registered into the
        given DNS server. (More likely local subjects).

            ::

                {
                    "hostname": "",
                    "aliases": [],
                    "ips": []
                }

        A dict with following index for everything else (and if found).

            ::

                {
                    "A": [],
                    "AAAA": [],
                    "CNAME": [],
                    "MX": [],
                    "NS": [],
                    "TXT": [],
                    "PTR": []
                }

    :rtype: dict
    """

    if subject:
        # The subject is not empty nor None.

        # We return the lookup.
        return lookup.Dns(dns_server=dns_server, lifetime=lifetime).request(
            subject, complete=complete
        )

    # We return None, there is nothing to work with.
    return None


def whois(subject, server=None, timeout=3):  # pragma: no cover
    """
    Request the WHOIS record of the given subject.

    :param str subject: The subject we are working with.
    :param str server:
        The WHOIS server to communicate with.

        .. note::
            If :code:`None` is given, we look for the best one.
    :param int timeout: The timeout to apply to the request.

    :return: None or the WHOIS record.
    :rtype: None|str
    """

    if subject:
        # The subject is not empty nor None.

        # We return the whois record.
        return lookup.Whois(subject, server=server, timeout=timeout).request()

    # We return None, there is nothing to work with.
    return None


def syntax_check(domain):  # pragma: no cover
    """
    Check the syntax of the given domain.

    :param domain: The domain to check the syntax from.
    :type domain: str|list

    :return: The syntax validity.
    :rtype: bool|dict

    .. warning::
        This method will be deprecated one day in the future.

        Please report to :func:`~PyFunceble.is_domain`.
    """

    warnings.warn(
        "`PyFunceble.syntax_check` will be deprecated in future version. "
        "Please use `PyFunceble.is_domain` instead.",
        DeprecationWarning,
    )

    return is_domain(domain)


def is_domain(subject):  # pragma: no cover
    """
    Check if the given subject is a syntactically valid domain.

    :param subject: The subject to check the syntax from.
    :type subject: str|list

    :return: The syntax validity.
    :rtype: bool|dict
    """

    if subject:
        # The given subject is not empty nor None.

        # We return the validiry of the given subject.
        return core.API(subject).domain_syntax()

    # We return None, there is nothing to check.
    return None


def is_subdomain(subject):  # pragma: no cover
    """
    Check if the given subject is a syntactically valid subdomain.

    :param subject: The subject to check the syntax from.
    :type subject: str|list

    :return: The syntax validity.
    :rtype: bool|dict
    """

    if subject:
        # The given subject is not empty nor None.

        # We retun the validity of the given subject.
        return core.API(subject).subdomain_syntax()

    # We return None, there is nothing to check.
    return None


def ipv4_syntax_check(ip):  # pragma: no cover
    """
    Check the syntax of the given IPv4.

    :param ip: The IPv4 to check the syntax for.
    :type ip: str|list

    :return: The syntax validity.
    :rtype: bool|dict

    .. warning::
        This method will be deprecated one day in the future.

        Please report to :func:`~PyFunceble.is_ipv4`.
    """

    warnings.warn(
        "`PyFunceble.ipv4_syntax_check` will be deprecated in future version. "
        "Please use `PyFunceble.is_ipv4` instead.",
        DeprecationWarning,
    )

    return is_ipv4(ip)


def is_ip(subject):  # pragma: no cover
    """
    Checks if the given subject is a syntactivally valid IPv4 or IPv6.

    :param subject: The subject to check the syntax from.
    :type subject: str|list

    :return: The syntax validity.
    :rtype: bool|dict
    """

    if subject:
        # The given subject is not empty nor None.

        # We return the validity of the given subject.
        return core.API(subject).ip_syntax()

    # We return None, there is nothing to check.
    return None


def is_ipv4(subject):  # pragma: no cover
    """
    Check if the given subject is a syntactically valid IPv4.

    :param subject: The subject to check the syntax from.
    :type subject: str|list

    :return: The syntax validity.
    :rtype: bool|dict
    """

    if subject:
        # The given subject is not empty nor None.

        # We return the validity of the given subject.
        return core.API(subject).ipv4_syntax()

    # We return None, there is nothing to check.
    return None


def is_ipv6(subject):  # pragma: no cover
    """
    Checks if the given subject is syntactivally valid IPv6.

    :param subject: The subject to check the syntax from.
    :type subject: str, list

    :return: The syntax validity.
    :rtype: bool|dict
    """

    if subject:
        # The given subject is not empty not None.

        # We return the validity of the given subject.
        return core.API(subject).ipv6_syntax()

    # We return None, there is nothing to check.
    return None


def is_ip_range(subject):  # pragma: no cover
    """
    Check if the given subject is a syntactically valid IPv4 or IPv6 range.

    :param subject: The subject to check the syntax from.
    :type subject: str|list

    :return: The IPv4 range state.
    :rtype: bool|dict
    """

    if subject:
        # The given subject is not empty nor None.

        # We return the validity of the given subject.
        return core.API(subject).ip_range_syntax()

    # We return None, there is nothing to check.
    return None


def is_ipv4_range(subject):  # pragma: no cover
    """
    Check if the given subject is a syntactically valid IPv4 range.

    :param subject: The subject to check the syntax from.
    :type subject: str|list

    :return: The IPv4 range state.
    :rtype: bool|dict
    """

    if subject:
        # The given subject is not empty nor None.

        # We return the validity of the given subject.
        return core.API(subject).ipv4_range_syntax()

    # We return None, there is nothing to check.
    return None


def url_syntax_check(url):  # pragma: no cover
    """
    Check the syntax of the given URL.

    :param url: The URL to check the syntax for.
    :type url: str|list

    :return: The syntax validity.
    :rtype: bool|dict

    .. warning::
        This method will be deprecated one day in the future.

        Please report to :func:`~PyFunceble.is_url`.
    """

    warnings.warn(
        "`PyFunceble.url_syntax_check` will be deprecated in future version. "
        "Please use `PyFunceble.is_url` instead.",
        DeprecationWarning,
    )

    return is_url(url)


def is_url(subject):  # pragma: no cover
    """
    Check if the given subject is a syntactically valid URL.

    :param subject: The subject to check the syntax from.
    :type subject: str|list

    :return: The syntax validity.
    :rtype: bool|dict
    """

    if subject:
        # The given subject is not empty nor None.

        # We return the validity of the given subject.
        return core.API(subject).url_syntax()

    # We return None, there is nothing to check.
    return None


def load_config(generate_directory_structure=False, custom=None):  # pragma: no cover
    """
    Load the configuration.

    :param bool generate_directory_structure:
        Tell us if we generate the directory structure
        along with loading the configuration file.

    :param dict custom:
        A dict with the configuration index (from .PyFunceble.yaml) to update.

    .. note::
        If :code:`config` is given, the given :code:`dict` overwrite
        the last value of the given indexes in the configuration.

        It's actually something like following:

        ::

            pyfunceble.configuration.update(config_given_by_user)
    """

    if not LOADER:
        loader = cconfig.Loader()
        loader.set_path_to_config(CONFIG_DIRECTORY)
        loader.get_config()
        loader.set_custom_config(custom)
    elif not LOADER.was_configuration_loaded():
        LOADER.set_path_to_config(CONFIG_DIRECTORY)
        LOADER.get_config()
        LOADER.set_custom_config(custom)
    else:
        LOADER.set_custom_config(custom)

    if generate_directory_structure:
        output.Constructor()


def is_domain_malicious(subject):  # pragma: no cover
    """
    Checks if the given domain is malicious.

    :param str subject: The subject to work with.

    :rtype: bool
    """

    if subject:
        return core.API(subject).reputation("domain") == "MALICIOUS"
    return None


def is_ipv4_malicious(subject):  # pragma: no cover
    """
    Checks if the given IPv4 is malicious.

    :rtype: bool
    """

    return is_domain_malicious(subject)


def is_url_malicious(subject):  # pragma: no cover
    """
    Checks if the given URL is malicious.

    :param str subject: The subject to work with.

    :rtype: bool
    """

    if subject:
        return core.API(subject).reputation("url") == "MALICIOUS"
    return None


def get_complements(subject, include_given=False):
    """
    Provides the complements of the given subject(s).

    A complement is for example :code:`example.com` if :code:`www.example.com`
    is given and vice-versa.

    :param subject: The subject to get the complement for.
    :type subject: str, list

    :param bool include_given:
        Tell us to add the given one into the result.

    :rtype: list
    """

    complements = []

    if isinstance(subject, str):
        checker = Check(subject)

        if include_given and subject not in complements:
            complements.append(subject)

        if subject.startswith("www."):
            complements.append(subject[4:])
        elif checker.is_domain() and not checker.is_subdomain():
            complements.append(f"www.{subject}")
    elif isinstance(subject, (list, set)):
        for subj in subject:
            complements.extend(get_complements(subj, include_given=include_given))

    return complements
